# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from abc import ABC, abstractmethod
from typing import Coroutine, List

from .._mars.core.context import Context, get_context
from .._mars.deploy.oscar.session import get_default_session
from .._mars.lib.aio import Isolation, new_isolation
from .._mars.services.cluster import ClusterAPI, NodeRole, WebClusterAPI
from .._mars.utils import implements


class API(ABC):
    """
    This is now for internal usage, APIs could change without notification.
    """

    @classmethod
    def create(cls):
        ctx = get_context()
        if ctx is not None:
            return _MarsContextBasedAPI(ctx)
        else:
            mars_session = get_default_session()
            if mars_session is None:
                raise ValueError("Xorbits is not inited yet, call `xorbits.init()`")
            return _ClientAPI(mars_session.address)

    @abstractmethod
    def list_workers(self) -> List[str]:
        """
        List addresses of workers.

        Returns
        -------
        addresses:
            Addresses of workers
        """


class _MarsContextBasedAPI(API):
    _mars_ctx: Context

    def __init__(self, mars_ctx: Context):
        self._mars_ctx = mars_ctx

    @implements(API.list_workers)
    def list_workers(self) -> List[str]:
        return self._mars_ctx.get_worker_addresses()


class _ClientAPI(API):
    _supervisor_address: str
    _isolation: Isolation
    _mars_cluster_api: ClusterAPI

    def __init__(self, supervisor_addr: str):
        self._supervisor_address = supervisor_addr

        self._inited = False
        # Isolation
        self._isolation = new_isolation("xorbits_api")
        # Mars service APIs
        self._mars_cluster_api = None

    async def _init(self):
        if self._inited:
            return

        if "://" in self._supervisor_address:
            self._mars_cluster_api = WebClusterAPI(self._supervisor_address)
        else:
            self._mars_cluster_api = await ClusterAPI.create(self._supervisor_address)
        self._inited = True

    def _call(self, coro: Coroutine):
        fut = asyncio.run_coroutine_threadsafe(coro, self._isolation.loop)
        return fut.result()

    async def _list_workers(self) -> List[str]:
        await self._init()
        return list(await self._mars_cluster_api.get_nodes_info(role=NodeRole.WORKER))

    @implements(API.list_workers)
    def list_workers(self) -> List[str]:
        return self._call(self._list_workers())



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import scipy.special as spspecial

from ..arithmetic.utils import arithmetic_operand
from ..utils import implement_scipy, infer_dtype
from .core import (
    TensorSpecialMultiOp,
    TensorSpecialUnaryOp,
    TensorTupleOp,
    _register_special_op,
)


@_register_special_op
@arithmetic_operand(sparse_mode="unary")
class TensorErf(TensorSpecialUnaryOp):
    _func_name = "erf"


@_register_special_op
@arithmetic_operand(sparse_mode="unary")
class TensorErfc(TensorSpecialUnaryOp):
    _func_name = "erfc"


@_register_special_op
@arithmetic_operand(sparse_mode="unary")
class TensorErfcx(TensorSpecialUnaryOp):
    _func_name = "erfcx"


@_register_special_op
@arithmetic_operand(sparse_mode="unary")
class TensorErfi(TensorSpecialUnaryOp):
    _func_name = "erfi"


@_register_special_op
@arithmetic_operand(sparse_mode="unary")
class TensorErfinv(TensorSpecialUnaryOp):
    _func_name = "erfinv"


@_register_special_op
@arithmetic_operand(sparse_mode="unary")
class TensorErfcinv(TensorSpecialUnaryOp):
    _func_name = "erfcinv"


@_register_special_op
@arithmetic_operand(sparse_mode="unary")
class TensorWofz(TensorSpecialUnaryOp):
    _func_name = "wofz"


@_register_special_op
@arithmetic_operand(sparse_mode="unary")
class TensorDawsn(TensorSpecialUnaryOp):
    _func_name = "dawsn"


@_register_special_op
class TensorFresnel(TensorTupleOp):
    _func_name = "fresnel"
    _n_outputs = 2


@_register_special_op
class TensorModFresnelP(TensorTupleOp):
    _func_name = "modfresnelp"
    _n_outputs = 2


@_register_special_op
class TensorModFresnelM(TensorTupleOp):
    _func_name = "modfresnelm"
    _n_outputs = 2


@_register_special_op
class TensorVoigtProfile(TensorSpecialMultiOp):
    _ARG_COUNT = 3
    _func_name = "voigt_profile"


@implement_scipy(spspecial.erf)
@infer_dtype(spspecial.erf)
def erf(x, out=None, where=None, **kwargs):
    """
    Returns the error function of complex argument.

    It is defined as ``2/sqrt(pi)*integral(exp(-t**2), t=0..z)``.

    Parameters
    ----------
    x : Tensor
        Input tensor.

    Returns
    -------
    res : Tensor
        The values of the error function at the given points `x`.

    See Also
    --------
    erfc, erfinv, erfcinv, wofz, erfcx, erfi

    Notes
    -----
    The cumulative of the unit normal distribution is given by
    ``Phi(z) = 1/2[1 + erf(z/sqrt(2))]``.

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Error_function
    .. [2] Milton Abramowitz and Irene A. Stegun, eds.
        Handbook of Mathematical Functions with Formulas,
        Graphs, and Mathematical Tables. New York: Dover,
        1972. http://www.math.sfu.ca/~cbm/aands/page_297.htm
    .. [3] Steven G. Johnson, Faddeeva W function implementation.
       http://ab-initio.mit.edu/Faddeeva

    Examples
    --------
    >>> import mars.tensor as mt
    >>> from mars.tensor import special
    >>> import matplotlib.pyplot as plt
    >>> x = mt.linspace(-3, 3)
    >>> plt.plot(x, special.erf(x))
    >>> plt.xlabel('$x$')
    >>> plt.ylabel('$erf(x)$')
    >>> plt.show()
    """
    op = TensorErf(**kwargs)
    return op(x, out=out, where=where)


@implement_scipy(spspecial.erfc)
@infer_dtype(spspecial.erfc)
def erfc(x, out=None, where=None, **kwargs):
    op = TensorErfc(**kwargs)
    return op(x, out=out, where=where)


@implement_scipy(spspecial.erfcx)
@infer_dtype(spspecial.erfcx)
def erfcx(x, out=None, where=None, **kwargs):
    op = TensorErfcx(**kwargs)
    return op(x, out=out, where=where)


@implement_scipy(spspecial.erfi)
@infer_dtype(spspecial.erfi)
def erfi(x, out=None, where=None, **kwargs):
    op = TensorErfi(**kwargs)
    return op(x, out=out, where=where)


@implement_scipy(spspecial.erfinv)
@infer_dtype(spspecial.erfinv)
def erfinv(x, out=None, where=None, **kwargs):
    op = TensorErfinv(**kwargs)
    return op(x, out=out, where=where)


@implement_scipy(spspecial.erfcinv)
@infer_dtype(spspecial.erfcinv)
def erfcinv(x, out=None, where=None, **kwargs):
    op = TensorErfcinv(**kwargs)
    return op(x, out=out, where=where)


@implement_scipy(spspecial.wofz)
@infer_dtype(spspecial.wofz)
def wofz(x, out=None, where=None, **kwargs):
    op = TensorWofz(**kwargs)
    return op(x, out=out, where=where)


@implement_scipy(spspecial.dawsn)
@infer_dtype(spspecial.dawsn)
def dawsn(x, out=None, where=None, **kwargs):
    op = TensorDawsn(**kwargs)
    return op(x, out=out, where=where)


@implement_scipy(spspecial.fresnel)
@infer_dtype(spspecial.fresnel, multi_outputs=True)
def fresnel(x, out=None, **kwargs):
    op = TensorFresnel(**kwargs)
    return op(x, out=out)


@implement_scipy(spspecial.modfresnelp)
@infer_dtype(spspecial.modfresnelp, multi_outputs=True)
def modfresnelp(x, out=None, **kwargs):
    op = TensorModFresnelP(**kwargs)
    return op(x, out=out)


@implement_scipy(spspecial.modfresnelm)
@infer_dtype(spspecial.modfresnelm, multi_outputs=True)
def modfresnelm(x, out=None, **kwargs):
    op = TensorModFresnelM(**kwargs)
    return op(x, out=out)


def voigt_profile(x, sigma, gamma, **kwargs):
    op = TensorVoigtProfile(**kwargs)
    return op(x, sigma, gamma)


if hasattr(spspecial, "voigt_profile"):
    func = spspecial.voigt_profile
    voigt_profile = implement_scipy(func)(infer_dtype(func)(voigt_profile))



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from typing import Any, Callable, Dict, Optional

from numpy import (
    NAN,
    NINF,
    AxisError,
    Inf,
    NaN,
    bool_,
    bytes_,
    cfloat,
    character,
    complex64,
    complex128,
    complexfloating,
    datetime64,
    double,
    dtype,
    e,
    errstate,
    finfo,
    flexible,
    float16,
    float32,
    float64,
    float_,
    floating,
    generic,
    inexact,
    inf,
    int8,
    int16,
    int32,
    int64,
    int_,
    intc,
    integer,
    intp,
    nan,
    newaxis,
    number,
    object_,
    pi,
    signedinteger,
    timedelta64,
    uint,
    uint8,
    uint16,
    uint32,
    uint64,
    unicode_,
    unsignedinteger,
    void,
)
from numpy.lib.index_tricks import ndindex

from ..core.utils.fallback import unimplemented_func


def _install():
    from .mars_adapters import _install as _install_mars_methods
    from .numpy_adapters import _install as _install_numpy_methods

    _install_mars_methods()
    _install_numpy_methods()


try:
    import warnings

    # suppress numpy warnings on types
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        warnings.simplefilter("ignore", FutureWarning)
        # noinspection PyUnresolvedReferences
        from numpy import bool, float, int, object  # type: ignore[attr-defined]
except ImportError:  # pragma: no cover
    pass
finally:
    del warnings


from . import fft, linalg, random, special
from .core import ndarray


def __dir__():
    from .mars_adapters import MARS_TENSOR_CALLABLES, MARS_TENSOR_OBJECTS
    from .numpy_adapters.core import NUMPY_MEMBERS

    return (
        list(MARS_TENSOR_CALLABLES.keys())
        + list((MARS_TENSOR_OBJECTS.keys()))
        + list(NUMPY_MEMBERS.keys())
    )


def __getattr__(name: str):
    from .mars_adapters import MARS_TENSOR_CALLABLES, MARS_TENSOR_OBJECTS
    from .numpy_adapters.core import NUMPY_MEMBERS

    if name in MARS_TENSOR_CALLABLES:
        return MARS_TENSOR_CALLABLES[name]
    elif name in MARS_TENSOR_OBJECTS:
        return MARS_TENSOR_OBJECTS[name]
    else:
        import numpy

        if not hasattr(numpy, name):
            raise AttributeError(name)
        elif name in NUMPY_MEMBERS:
            return NUMPY_MEMBERS[name]
        else:  # pragma: no cover
            if inspect.ismethod(getattr(numpy, name)):
                return unimplemented_func
            else:
                raise AttributeError(name)



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from xoscar.backends.router import Router

from . import numpy as np
from . import pandas as pd
from ._mars.config import option_context
from ._mars.deploy.oscar.session import clear_default_session
from .tests.core import init_test


def pytest_addoption(parser):
    parser.addoption(
        "--gpu", action="store_true", default=False, help="run test cases on GPU."
    )


@pytest.fixture
def gpu(request):
    return request.config.option.gpu


@pytest.fixture
def doctest_namespace():
    return {"pd": pd, "np": np}


@pytest.fixture
def empty_df():
    return pd.DataFrame()


@pytest.fixture
def dummy_df():
    return pd.DataFrame({"foo": (0, 1, 2), "bar": ("a", "b", "c")})


@pytest.fixture
def dummy_int_series():
    return pd.Series([1, 2, 3, 4, 5])


@pytest.fixture
def dummy_str_series():
    return pd.Series(["foo", "bar", "baz"])


@pytest.fixture
def dummy_dt_series():
    return pd.Series(pd.date_range("2000-01-01", periods=3, freq="s"))


@pytest.fixture
def dummy_int_1d_array():
    return np.array([0, 1, 2])


@pytest.fixture
def dummy_int_2d_array():
    return np.arange(9).reshape(3, 3)


@pytest.fixture(scope="package")
def _setup_test_session():
    sess = init_test(
        address="test://127.0.0.1",
        backend="mars",
        init_local=True,
        default=True,
        web=False,
        timeout=300,
    )
    with option_context({"show_progress": False}):
        try:
            yield sess
        finally:
            sess.stop_server(isolation=False)
            Router.set_instance(None)


@pytest.fixture
def setup(_setup_test_session):
    _setup_test_session.as_default()
    yield _setup_test_session
    clear_default_session()


@pytest.fixture
def _setup_with_web():
    sess = init_test(web=True)
    try:
        yield sess
    finally:
        sess.stop_server(isolation=False)
        Router.set_instance(None)


@pytest.fixture
def setup_with_web(_setup_with_web):
    _setup_with_web.as_default()
    yield _setup_with_web
    clear_default_session()



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
from dataclasses import dataclass
from io import UnsupportedOperation
from typing import Any, Dict, List, Union

import xoscar as mo
from xoscar.core import BufferRef

from ...lib.aio import alru_cache
from ...storage import StorageLevel
from ...utils import dataslots
from .core import DataManagerActor, WrappedStorageFileObject
from .handler import StorageHandlerActor

DEFAULT_TRANSFER_BLOCK_SIZE = 4 * 1024**2


logger = logging.getLogger(__name__)


class SenderManagerActor(mo.StatelessActor):
    def __init__(
        self,
        band_name: str = "numa-0",
        transfer_block_size: int = None,
        data_manager_ref: mo.ActorRefType[DataManagerActor] = None,
        storage_handler_ref: mo.ActorRefType[StorageHandlerActor] = None,
    ):
        self._band_name = band_name
        self._data_manager_ref = data_manager_ref
        self._storage_handler = storage_handler_ref
        self._transfer_block_size = transfer_block_size or DEFAULT_TRANSFER_BLOCK_SIZE

    @classmethod
    def gen_uid(cls, band_name: str):
        return f"sender_manager_{band_name}"

    async def __post_create__(self):
        if self._storage_handler is None:  # for test
            self._storage_handler = await mo.actor_ref(
                self.address, StorageHandlerActor.gen_uid("numa-0")
            )

    @staticmethod
    @alru_cache
    async def get_receiver_ref(address: str, band_name: str):
        return await mo.actor_ref(
            address=address, uid=ReceiverManagerActor.gen_uid(band_name)
        )

    async def _copy_to_receiver(
        self,
        receiver_ref: mo.ActorRefType["ReceiverManagerActor"],
        local_buffers: List,
        remote_buffers: List[BufferRef],
        session_id: str,
        data_keys: List[str],
        block_size: int,
    ):
        await mo.copy_to(local_buffers, remote_buffers, block_size=block_size)
        await receiver_ref.handle_transmission_done(session_id, data_keys)

    @staticmethod
    def _get_buffer_size(buf: Any):
        if hasattr(buf, "size"):
            return buf.size
        return len(buf)

    async def _send_data(
        self,
        receiver_ref: mo.ActorRefType["ReceiverManagerActor"],
        session_id: str,
        data_keys: List[str],
        block_size: int,
    ):
        open_reader_tasks = []
        for data_key in data_keys:
            open_reader_tasks.append(
                self._storage_handler.open_reader.delay(session_id, data_key)
            )
        readers = await self._storage_handler.open_reader.batch(*open_reader_tasks)

        local_buffers = []
        data_sizes = []
        headers = []
        for reader in readers:
            try:
                reader_buffer = reader.buffer
            except UnsupportedOperation:  # vineyard  # pragma: no cover
                logger.warning(
                    "Transfer via buffer not work for vineyard. Use file object instead."
                )
                reader_buffer = None

            if reader_buffer is not None:
                if isinstance(reader_buffer, list):  # cuda case
                    data_sizes.append(
                        [self._get_buffer_size(rb) for rb in reader_buffer]
                    )
                    local_buffers.extend(reader_buffer)
                    headers.append(reader.header)
                else:
                    data_sizes.append(
                        getattr(reader_buffer, "size", len(reader_buffer))
                    )
                    local_buffers.append(reader_buffer)
                    headers.append(None)
            else:  # file system
                data_sizes.append(None)
                local_buffers.append(reader)
                headers.append(None)

        remote_buffers = await receiver_ref.get_buffers(
            session_id, data_keys, data_sizes, headers
        )

        write_task = asyncio.create_task(
            self._copy_to_receiver(
                receiver_ref,
                local_buffers,
                remote_buffers,
                session_id,
                data_keys,
                block_size,
            )
        )

        try:
            await asyncio.shield(write_task)
        except asyncio.CancelledError:
            await receiver_ref.handle_transmission_cancellation(session_id, data_keys)
            raise

    @mo.extensible
    async def send_batch_data(
        self,
        session_id: str,
        data_keys: List[str],
        address: str,
        level: StorageLevel,
        band_name: str = "numa-0",
        block_size: int = None,
        error: str = "raise",
    ):
        logger.debug(
            "Begin to send data (%s, %s) to %s", session_id, data_keys, address
        )

        tasks = []
        for key in data_keys:
            tasks.append(self._data_manager_ref.get_store_key.delay(session_id, key))
        data_keys = await self._data_manager_ref.get_store_key.batch(*tasks)
        data_keys = list(set(data_keys))
        sub_infos = await self._data_manager_ref.get_sub_infos.batch(
            *[
                self._data_manager_ref.get_sub_infos.delay(session_id, key)
                for key in data_keys
            ]
        )

        block_size = block_size or self._transfer_block_size
        receiver_ref: mo.ActorRefType[
            ReceiverManagerActor
        ] = await self.get_receiver_ref(address, band_name)
        get_infos = []
        pin_tasks = []
        for data_key in data_keys:
            get_infos.append(
                self._data_manager_ref.get_data_info.delay(
                    session_id, data_key, self._band_name, error
                )
            )
            pin_tasks.append(
                self._data_manager_ref.pin.delay(
                    session_id, data_key, self._band_name, error
                )
            )
        await self._data_manager_ref.pin.batch(*pin_tasks)
        infos = await self._data_manager_ref.get_data_info.batch(*get_infos)
        filtered = [
            (data_info, data_key)
            for data_info, data_key in zip(infos, data_keys)
            if data_info is not None
        ]
        if filtered:
            infos, data_keys = zip(*filtered)
        else:  # pragma: no cover
            # no data to be transferred
            return
        data_sizes = [info.store_size for info in infos]
        if level is None:
            level = infos[0].level
        is_transferring_list = await receiver_ref.open_writers(
            session_id, data_keys, data_sizes, level, sub_infos, band_name
        )
        to_send_keys = []
        to_wait_keys = []
        for data_key, is_transferring in zip(data_keys, is_transferring_list):
            if is_transferring:
                to_wait_keys.append(data_key)
            else:
                to_send_keys.append(data_key)

        if to_send_keys:
            await self._send_data(receiver_ref, session_id, to_send_keys, block_size)
        if to_wait_keys:
            await receiver_ref.wait_transfer_done(session_id, to_wait_keys)
        unpin_tasks = []
        for data_key in data_keys:
            unpin_tasks.append(
                self._data_manager_ref.unpin.delay(
                    session_id, [data_key], self._band_name, error="ignore"
                )
            )
        await self._data_manager_ref.unpin.batch(*unpin_tasks)
        logger.debug(
            "Finish sending data (%s, %s) to %s, total size is %s",
            session_id,
            data_keys,
            address,
            sum(data_sizes),
        )


@dataslots
@dataclass
class WritingInfo:
    writer: WrappedStorageFileObject
    size: int
    level: StorageLevel
    event: asyncio.Event
    ref_counts: int


class ReceiverManagerActor(mo.StatelessActor):
    def __init__(
        self,
        quota_refs: Dict,
        storage_handler_ref: mo.ActorRefType[StorageHandlerActor] = None,
    ):
        self._quota_refs = quota_refs
        self._storage_handler = storage_handler_ref
        self._writing_infos: Dict[tuple, WritingInfo] = dict()
        self._lock = asyncio.Lock()

    async def __post_create__(self):
        if self._storage_handler is None:  # for test
            self._storage_handler = await mo.actor_ref(
                self.address, StorageHandlerActor.gen_uid("numa-0")
            )

    async def get_buffers(
        self,
        session_id: str,
        data_keys: List,
        data_sizes: List[Union[int, List[int]]],
        headers: List,
    ) -> List:
        res = []
        for data_key, data_size, header in zip(data_keys, data_sizes, headers):
            writer = self._writing_infos[(session_id, data_key)].writer
            if isinstance(data_size, list):  # cuda case
                writer.set_buffers_by_sizes(data_size)
            if header:
                writer.set_file_header(header)

            try:
                writer_buf = writer.buffer
            except UnsupportedOperation:  # pragma: no cover
                writer_buf = None
            if writer_buf is not None:
                if isinstance(writer_buf, list):  # cuda case
                    res.extend([mo.buffer_ref(self.address, buf) for buf in writer_buf])
                else:
                    res.append(mo.buffer_ref(self.address, writer_buf))
            else:  # file system
                res.append(mo.file_object_ref(self.address, writer))
        return res

    async def handle_transmission_done(self, session_id: str, data_keys: List):
        close_tasks = [
            self._writing_infos[(session_id, data_key)].writer.close()
            for data_key in data_keys
        ]
        await asyncio.gather(*close_tasks)
        async with self._lock:
            for data_key in data_keys:
                event = self._writing_infos[(session_id, data_key)].event
                event.set()
                self._decref_writing_key(session_id, data_key)

    async def handle_transmission_cancellation(self, session_id: str, data_keys: List):
        async with self._lock:
            for data_key in data_keys:
                if (session_id, data_key) in self._writing_infos:
                    if self._writing_infos[(session_id, data_key)].ref_counts == 1:
                        info = self._writing_infos[(session_id, data_key)]
                        await self._quota_refs[info.level].release_quota(info.size)
                        await self._storage_handler.delete(
                            session_id, data_key, error="ignore"
                        )
                        await info.writer.clean_up()
                        info.event.set()
                        self._decref_writing_key(session_id, data_key)

    @classmethod
    def gen_uid(cls, band_name: str):
        return f"receiver_manager_{band_name}"

    def _decref_writing_key(self, session_id: str, data_key: str):
        self._writing_infos[(session_id, data_key)].ref_counts -= 1
        if self._writing_infos[(session_id, data_key)].ref_counts == 0:
            del self._writing_infos[(session_id, data_key)]

    async def create_writers(
        self,
        session_id: str,
        data_keys: List[str],
        data_sizes: List[int],
        level: StorageLevel,
        sub_infos: List,
        band_name: str,
    ):
        tasks = dict()
        key_to_sub_infos = dict()
        data_key_to_size = dict()
        being_processed = []
        for data_key, data_size, sub_info in zip(data_keys, data_sizes, sub_infos):
            data_key_to_size[data_key] = data_size
            if (session_id, data_key) not in self._writing_infos:
                being_processed.append(False)
                tasks[data_key] = self._storage_handler.open_writer.delay(
                    session_id,
                    data_key,
                    data_size,
                    level,
                    request_quota=False,
                    band_name=band_name,
                )
                key_to_sub_infos[data_key] = sub_info
            else:
                being_processed.append(True)
                self._writing_infos[(session_id, data_key)].ref_counts += 1
        if tasks:
            writers = await self._storage_handler.open_writer.batch(
                *tuple(tasks.values())
            )
            for data_key, writer in zip(tasks, writers):
                self._writing_infos[(session_id, data_key)] = WritingInfo(
                    writer, data_key_to_size[data_key], level, asyncio.Event(), 1
                )
                if key_to_sub_infos[data_key] is not None:
                    writer._sub_key_infos = key_to_sub_infos[data_key]
        return being_processed

    async def open_writers(
        self,
        session_id: str,
        data_keys: List[str],
        data_sizes: List[int],
        level: StorageLevel,
        sub_infos: List,
        band_name: str,
    ):
        async with self._lock:
            await self._storage_handler.request_quota_with_spill(level, sum(data_sizes))
            future = asyncio.create_task(
                self.create_writers(
                    session_id, data_keys, data_sizes, level, sub_infos, band_name
                )
            )
            try:
                return await future
            except asyncio.CancelledError:
                await self._quota_refs[level].release_quota(sum(data_sizes))
                future.cancel()
                raise

    async def wait_transfer_done(self, session_id, data_keys):
        await asyncio.gather(
            *[self._writing_infos[(session_id, key)].event.wait() for key in data_keys]
        )
        async with self._lock:
            for data_key in data_keys:
                self._decref_writing_key(session_id, data_key)



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import pytest

from .... import pandas as xpd

df_raw = pd.DataFrame(
    {
        "A": [1.0] * 100,
        "B": [2.0] * 100,
        "C": pd.date_range(start="2023-06-19", periods=100),
    }
)
df = xpd.DataFrame(df_raw)


def test_options_execute(setup):
    # test pandas options
    pd.set_option("display.max_rows", 100)
    xpd.set_option("display.max_rows", 100)
    assert (
        pd.get_option("display.max_rows") == xpd.get_option("display.max_rows") == 100
    )
    assert df_raw.__str__() == df.__str__()

    # test xorbits options
    xpd.set_option("chunk_store_limit", 100000)
    assert xpd.get_option("chunk_store_limit") == 100000

    # test reset options
    xpd.reset_option("chunk_store_limit")
    assert xpd.get_option("chunk_store_limit") == 134217728
    xpd.reset_option("display.max_rows")
    assert xpd.get_option("display.max_rows") == 60

    # test option_context
    with xpd.option_context("display.max_rows", 100, "chunk_store_limit", 100000):
        assert xpd.get_option("display.max_rows") == 100
        assert xpd.get_option("chunk_store_limit") == 100000

    # test set_eng_float_format
    pd.set_eng_float_format(accuracy=3)
    xpd.set_eng_float_format(accuracy=3)
    assert df_raw.__str__() == df.__str__()

    # test error value
    with pytest.raises(ValueError):
        xpd.set_option("display.max_rows", "100")

    # test error option
    with pytest.raises(pd._config.config.OptionError):
        xpd.set_option("non-exist", None)

    with pytest.raises(pd._config.config.OptionError):
        xpd.get_option("non-exist")

    with pytest.raises(pd._config.config.OptionError):
        xpd.reset_option("non-exist")

    with pytest.raises(ValueError):
        xpd.option_context("non-exist", 100)

    # test invalid type
    with pytest.raises(ValueError):
        xpd.option_context({"display.max_rows": 100})


def test_docstring(setup):
    docstring = xpd.set_option.__doc__
    assert docstring is not None and docstring.endswith(
        "This docstring was copied from pandas."
    )

    docstring = xpd.get_option.__doc__
    assert docstring is not None and docstring.endswith(
        "This docstring was copied from pandas."
    )

    docstring = xpd.reset_option.__doc__
    assert docstring is not None and docstring.endswith(
        "This docstring was copied from pandas."
    )

    docstring = xpd.option_context.__doc__
    assert docstring is not None and docstring.endswith(
        "This docstring was copied from pandas."
    )

    docstring = xpd.set_eng_float_format.__doc__
    assert docstring is not None and docstring.endswith(
        "This docstring was copied from pandas."
    )



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import functools
import itertools
from collections import OrderedDict
from collections.abc import Iterable
from typing import Callable, Dict, List, Union

import numpy as np
import pandas as pd

from ... import opcodes
from ... import tensor as mars_tensor
from ...config import options
from ...core import ENTITY_TYPE, OutputType, enter_mode, recursive_tile
from ...core.custom_log import redirect_custom_log
from ...core.operand import OperandStage
from ...serialization.serializables import (
    AnyField,
    BoolField,
    DictField,
    Int32Field,
    ListField,
)
from ...utils import ceildiv, enter_current_session, lazy_import, pd_release_version
from ..core import INDEX_CHUNK_TYPE
from ..merge import DataFrameConcat
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import (
    build_df,
    build_empty_df,
    build_series,
    is_cudf,
    is_dataframe,
    is_series,
    parse_index,
    validate_axis,
)
from .core import CustomReduction, ReductionAggStep, ReductionCompiler, ReductionSteps

cp = lazy_import("cupy", rename="cp")
cudf = lazy_import("cudf")

_agg_size_as_series = pd_release_version >= (1, 3, 0)


def where_function(cond, var1, var2):
    if hasattr(var1, "ndim") and var1.ndim >= 1:
        return var1.where(cond, var2)
    elif isinstance(var1, ENTITY_TYPE):
        return mars_tensor.where(cond, var1, var2)
    else:
        return np.where(cond, var1, var2).item()


_agg_functions = {
    "sum": lambda x, skipna=True: x.sum(skipna=skipna),
    "prod": lambda x, skipna=True: x.prod(skipna=skipna),
    "product": lambda x, skipna=True: x.product(skipna=skipna),
    "min": lambda x, skipna=True: x.min(skipna=skipna),
    "max": lambda x, skipna=True: x.max(skipna=skipna),
    "all": lambda x, skipna=True: x.all(skipna=skipna),
    "any": lambda x, skipna=True: x.any(skipna=skipna),
    "count": lambda x: x.count(),
    "size": lambda x: x._reduction_size(),
    "mean": lambda x, skipna=True: x.mean(skipna=skipna),
    "var": lambda x, skipna=True, ddof=1: x.var(skipna=skipna, ddof=ddof),
    "std": lambda x, skipna=True, ddof=1: x.std(skipna=skipna, ddof=ddof),
    "sem": lambda x, skipna=True, ddof=1: x.sem(skipna=skipna, ddof=ddof),
    "skew": lambda x, skipna=True, bias=False: x.skew(skipna=skipna, bias=bias),
    "kurt": lambda x, skipna=True, bias=False: x.kurt(skipna=skipna, bias=bias),
    "kurtosis": lambda x, skipna=True, bias=False: x.kurtosis(skipna=skipna, bias=bias),
    "nunique": lambda x: x.reduction_nunique(),
}


class DataFrameAggregate(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.AGGREGATE

    raw_func = AnyField("raw_func")
    raw_func_kw = DictField("raw_func_kw")
    func = AnyField("func")
    func_rename = ListField("func_rename")
    axis = AnyField("axis")
    numeric_only = BoolField("numeric_only")
    bool_only = BoolField("bool_only")

    combine_size = Int32Field("combine_size")
    pre_funcs = ListField("pre_funcs")
    agg_funcs = ListField("agg_funcs")
    post_funcs = ListField("post_funcs")

    @staticmethod
    def _filter_dtypes(op: "DataFrameAggregate", dtypes):
        if not op.numeric_only and not op.bool_only:
            return dtypes
        empty_df = build_empty_df(dtypes)
        return empty_df.select_dtypes(
            [np.number, np.bool_] if op.numeric_only else [np.bool_]
        ).dtypes

    def _calc_result_shape(self, df):
        if df.ndim == 2:
            if self.numeric_only:
                df = df.select_dtypes([np.number, np.bool_])
            elif self.bool_only:
                df = df.select_dtypes([np.bool_])

        if self.output_types[0] == OutputType.dataframe:
            test_obj = build_df(df, size=[2, 2], fill_value=[1, 2], ensure_string=True)
        else:
            test_obj = build_series(
                df, size=[2, 2], fill_value=[1, 2], name=df.name, ensure_string=True
            )

        result_df = test_obj.agg(self.raw_func, axis=self.axis, **self.raw_func_kw)

        if isinstance(result_df, pd.DataFrame):
            self.output_types = [OutputType.dataframe]
            return result_df.dtypes, result_df.index
        elif isinstance(result_df, pd.Series):
            self.output_types = [OutputType.series]
            return pd.Series([result_df.dtype], index=[result_df.name]), result_df.index
        else:
            self.output_types = [OutputType.scalar]
            return np.array(result_df).dtype, None

    def __call__(self, df, output_type=None, dtypes=None, index=None):
        self._output_types = df.op.output_types
        normalize_reduction_funcs(self, ndim=df.ndim)
        if output_type is None or dtypes is None:
            with enter_mode(kernel=False, build=False):
                dtypes, index = self._calc_result_shape(df)
        else:
            self.output_types = [output_type]

        if self.output_types[0] == OutputType.dataframe:
            if self.axis == 0:
                new_shape = (len(index), len(dtypes))
                new_index = parse_index(index, store_data=True)
            else:
                new_shape = (df.shape[0], len(dtypes))
                new_index = df.index_value
            return self.new_dataframe(
                [df],
                shape=new_shape,
                dtypes=dtypes,
                index_value=new_index,
                columns_value=parse_index(dtypes.index, store_data=True),
            )
        elif self.output_types[0] == OutputType.series:
            if df.ndim == 1:
                new_shape = (len(index),)
                new_index = parse_index(index, store_data=True)
            elif self.axis == 0:
                new_shape = (len(index),)
                new_index = parse_index(index, store_data=True)
            else:
                new_shape = (df.shape[0],)
                new_index = df.index_value
            return self.new_series(
                [df],
                shape=new_shape,
                dtype=dtypes.iloc[0],
                name=dtypes.index[0],
                index_value=new_index,
            )
        elif self.output_types[0] == OutputType.tensor:
            return self.new_tileable([df], dtype=dtypes, shape=(np.nan,))
        else:
            return self.new_scalar([df], dtype=dtypes)

    @staticmethod
    def _safe_append(d, key, val):
        if key not in d:
            d[key] = []
        if val not in d[key]:
            d[key].append(val)

    @classmethod
    def _gen_map_chunks(
        cls,
        op,
        in_df,
        out_df,
        func_infos: List[ReductionSteps],
        input_index_to_output: Dict[int, int],
    ):
        axis = op.axis

        if axis == 0:
            agg_chunks_shape = (
                (in_df.chunk_shape[0], len(func_infos))
                if len(in_df.chunk_shape) == 2
                else (in_df.chunk_shape[0], 1)
            )
        else:
            agg_chunks_shape = (len(func_infos), in_df.chunk_shape[1])

        agg_chunks = np.empty(agg_chunks_shape, dtype=object)
        dtypes_cache = dict()
        for chunk in in_df.chunks:
            input_index = chunk.index[1 - axis] if len(chunk.index) > 1 else 0
            if input_index not in input_index_to_output:
                continue
            map_op = op.copy().reset_key()  # type: "DataFrameAggregate"
            new_axis_index = input_index_to_output[input_index]
            func_info = func_infos[new_axis_index]
            # force as_index=True for map phase
            map_op.output_types = (
                [OutputType.dataframe] if chunk.ndim == 2 else [OutputType.series]
            )
            map_op.stage = OperandStage.map
            map_op.pre_funcs = func_info.pre_funcs
            map_op.agg_funcs = func_info.agg_funcs

            if axis == 0:
                new_index = (
                    (chunk.index[0], new_axis_index)
                    if len(chunk.index) == 2
                    else (chunk.index[0], 0)
                )
            else:
                new_index = (new_axis_index, chunk.index[1])

            if map_op.output_types[0] == OutputType.dataframe:
                if axis == 0:
                    shape = (1, out_df.shape[-1])
                    if out_df.ndim == 2:
                        columns_value = out_df.columns_value
                        index_value = out_df.index_value
                    else:
                        columns_value = out_df.index_value
                        index_value = parse_index(pd.Index([0]), out_df.key)

                    try:
                        dtypes = dtypes_cache[chunk.index[1]]
                    except KeyError:
                        dtypes = chunk.dtypes.reindex(
                            columns_value.to_pandas()
                        ).dropna()
                        dtypes_cache[chunk.index[1]] = dtypes

                    agg_chunk = map_op.new_chunk(
                        [chunk],
                        shape=shape,
                        index=new_index,
                        dtypes=dtypes,
                        columns_value=columns_value,
                        index_value=index_value,
                    )
                else:
                    shape = (out_df.shape[0], 1)
                    columns_value = parse_index(
                        pd.Index([0]), out_df.key, store_data=True
                    )
                    index_value = out_df.index_value

                    agg_chunk = map_op.new_chunk(
                        [chunk],
                        shape=shape,
                        index=new_index,
                        columns_value=columns_value,
                        index_value=index_value,
                    )
            else:
                agg_chunk = map_op.new_chunk([chunk], shape=(1,), index=new_index)
            agg_chunks[agg_chunk.index] = agg_chunk
        return agg_chunks

    @classmethod
    def _tile_single_chunk(cls, op: "DataFrameAggregate"):
        in_df = op.inputs[0]
        out_df = op.outputs[0]

        chunk_op = op.copy().reset_key()
        if op.output_types[0] == OutputType.dataframe:
            chunk = chunk_op.new_chunk(
                in_df.chunks,
                index=(0, 0),
                shape=out_df.shape,
                index_value=out_df.index_value,
                columns_value=out_df.columns_value,
                dtypes=out_df.dtypes,
            )
        elif op.output_types[0] == OutputType.series:
            chunk = chunk_op.new_chunk(
                in_df.chunks,
                index=(0,),
                shape=out_df.shape,
                dtype=out_df.dtype,
                index_value=out_df.index_value,
                name=out_df.name,
            )
        elif op.output_types[0] == OutputType.tensor:
            chunk = chunk_op.new_chunk(
                in_df.chunks, index=(0,), dtype=out_df.dtype, shape=(np.nan,)
            )
        else:
            chunk = chunk_op.new_chunk(
                in_df.chunks, dtype=out_df.dtype, index=(), shape=()
            )

        tileable_op = op.copy().reset_key()
        kw = out_df.params.copy()
        kw.update(dict(chunks=[chunk], nsplits=tuple((x,) for x in out_df.shape)))
        return tileable_op.new_tileables([in_df], **kw)

    @classmethod
    def _tile_size(cls, op: "DataFrameAggregate"):
        in_df = op.inputs[0]
        out_df = op.outputs[0]

        chunks = []
        for c in in_df.chunks:
            chunk_op = op.copy().reset_key()
            chunks.append(
                chunk_op.new_chunk(
                    [c],
                    index=c.index,
                    shape=(1,) * len(in_df.shape),
                    dtype=out_df.dtype,
                )
            )

        tileable_op = op.copy().reset_key()
        nsplits = tuple((1,) * s for s in in_df.chunk_shape)
        tileable = tileable_op.new_tileable(
            out_df.inputs,
            chunks=chunks,
            nsplits=nsplits,
            shape=in_df.chunk_shape,
            dtype=out_df.dtype,
        )
        ret = yield from recursive_tile(tileable.sum())
        return [ret]

    @staticmethod
    def _add_functions(
        op: "DataFrameAggregate", compiler: ReductionCompiler, cols=None
    ):
        if isinstance(op.func, list):
            func_iter = ((None, f) for f in op.func)
            cols_set = set(cols) if cols is not None else None
        else:
            assert cols is not None
            cols_set = set(cols) & set(op.func.keys())
            if len(cols_set) == 0:
                return False
            func_iter = ((col, f) for col, funcs in op.func.items() for f in funcs)

        func_renames = (
            op.func_rename
            if getattr(op, "func_rename", None) is not None
            else itertools.repeat(None)
        )
        for func_rename, (col, f) in zip(func_renames, func_iter):
            if cols_set is not None and col is not None and col not in cols_set:
                continue
            func_name = None
            if isinstance(f, str):
                f, func_name = _agg_functions[f], f
            if func_rename is not None:
                func_name = func_rename
            ndim = 1 if cols is None else 2
            func_cols = [col] if col is not None else None
            compiler.add_function(f, ndim, cols=func_cols, func_name=func_name)
        return True

    @classmethod
    def _tile_tree(cls, op: "DataFrameAggregate"):
        in_df = op.inputs[0]
        out_df = op.outputs[0]
        combine_size = op.combine_size
        axis = op.axis

        input_index_to_output = dict()
        output_index_to_input = []
        axis_func_infos = []
        dtypes_list = []
        if len(in_df.chunk_shape) > 1:
            for col_idx in range(in_df.chunk_shape[1 - axis]):
                compiler = ReductionCompiler(axis=op.axis)
                idx_chunk = (
                    in_df.cix[0, col_idx] if axis == 0 else in_df.cix[col_idx, 0]
                )
                new_dtypes = cls._filter_dtypes(op, idx_chunk.dtypes)
                if not cls._add_functions(op, compiler, cols=list(new_dtypes.index)):
                    continue
                input_index_to_output[col_idx] = len(axis_func_infos)
                output_index_to_input.append(col_idx)
                axis_func_infos.append(compiler.compile())
                dtypes_list.append(new_dtypes)
        else:
            compiler = ReductionCompiler(axis=op.axis)
            cls._add_functions(op, compiler)
            input_index_to_output[0] = 0
            axis_func_infos.append(compiler.compile())

        chunks = cls._gen_map_chunks(
            op, in_df, out_df, axis_func_infos, input_index_to_output
        )
        while chunks.shape[axis] > combine_size:
            if axis == 0:
                new_chunks_shape = (
                    ceildiv(chunks.shape[0], combine_size),
                    chunks.shape[1],
                )
            else:
                new_chunks_shape = (
                    chunks.shape[0],
                    ceildiv(chunks.shape[1], combine_size),
                )

            new_chunks = np.empty(new_chunks_shape, dtype=object)
            for idx0, i in enumerate(range(0, chunks.shape[axis], combine_size)):
                for idx1 in range(chunks.shape[1 - axis]):
                    func_info = axis_func_infos[idx1]
                    if axis == 0:
                        chks = chunks[i : i + combine_size, idx1]
                        chunk_index = (idx0, idx1)
                        if chks[0].ndim == 1:
                            concat_shape = (len(chks),)
                            agg_shape = (1,)
                        else:
                            concat_shape = (len(chks), chks[0].shape[1])
                            agg_shape = (chks[0].shape[1], 1)
                    else:
                        chks = chunks[idx1, i : i + combine_size]
                        chunk_index = (idx1, idx0)
                        concat_shape = (chks[0].shape[0], len(chks))
                        agg_shape = (chks[0].shape[0], 1)

                    chks = chks.reshape((chks.shape[0],)).tolist()
                    if len(chks) == 1:
                        chk = chks[0]
                    else:
                        concat_op = DataFrameConcat(
                            output_types=[OutputType.dataframe], axis=axis
                        )
                        # Change index for concatenate
                        for j, c in enumerate(chks):
                            c._index = (j, 0) if axis == 0 else (0, j)
                        chk = concat_op.new_chunk(
                            chks,
                            dtypes=dtypes_list[idx1] if dtypes_list else None,
                            shape=concat_shape,
                            index_value=chks[0].index_value,
                        )
                    chunk_op = op.copy().reset_key()
                    chunk_op.output_types = [OutputType.dataframe]
                    chunk_op.stage = OperandStage.combine
                    chunk_op.agg_funcs = func_info.agg_funcs

                    if axis == 0:
                        new_chunks[chunk_index] = chunk_op.new_chunk(
                            [chk],
                            index=chunk_index,
                            shape=agg_shape,
                            index_value=chks[0].index_value,
                        )
                    else:
                        new_chunks[chunk_index] = chunk_op.new_chunk(
                            [chk],
                            index=chunk_index,
                            shape=agg_shape,
                            index_value=chks[0].columns_value,
                        )
            chunks = new_chunks

        agg_chunks = []
        for idx in range(chunks.shape[1 - axis]):
            func_info = axis_func_infos[idx]

            concat_op = DataFrameConcat(output_types=[OutputType.dataframe], axis=axis)
            if axis == 0:
                chks = chunks[:, idx]
                if chks[0].ndim == 1:
                    concat_shape = (len(chks),)
                else:
                    concat_shape = (len(chks), chks[0].shape[1])
            else:
                chks = chunks[idx, :]
                concat_shape = (chks[0].shape[0], len(chks))
            chks = chks.reshape((chks.shape[0],)).tolist()
            chk = concat_op.new_chunk(
                chks,
                dtypes=dtypes_list[idx] if dtypes_list else None,
                shape=concat_shape,
                index_value=chks[0].index_value,
            )
            chunk_op = op.copy().reset_key()
            chunk_op.stage = OperandStage.agg
            chunk_op.agg_funcs = func_info.agg_funcs
            chunk_op.post_funcs = func_info.post_funcs

            kw = out_df.params.copy()
            if op.output_types[0] == OutputType.dataframe:
                if axis == 0:
                    src_col_chunk = in_df.cix[0, output_index_to_input[idx]]
                    valid_cols = [
                        c for pre in func_info.pre_funcs for c in pre.columns or ()
                    ]
                    if not valid_cols:
                        columns_value = src_col_chunk.columns_value
                        shape_len = src_col_chunk.shape[1]
                    else:
                        col_index = pd.Index(valid_cols).unique()
                        columns_value = parse_index(col_index, store_data=True)
                        shape_len = len(col_index)
                    kw.update(
                        dict(
                            shape=(out_df.shape[0], shape_len),
                            columns_value=columns_value,
                            index=(0, idx),
                            dtypes=out_df.dtypes[columns_value.to_pandas()],
                        )
                    )
                else:
                    src_col_chunk = in_df.cix[output_index_to_input[idx], 0]
                    kw.update(
                        dict(
                            index=(idx, 0),
                            index_value=src_col_chunk.index_value,
                            shape=(src_col_chunk.shape[0], out_df.shape[1]),
                            dtypes=out_df.dtypes,
                        )
                    )
            else:
                if op.output_types[0] == OutputType.series:
                    if in_df.ndim == 1:
                        index_value, shape = out_df.index_value, out_df.shape
                    elif axis == 0:
                        out_dtypes = dtypes_list[idx]
                        index_value = parse_index(out_dtypes.index, store_data=True)
                        shape = (len(out_dtypes),)
                    else:
                        src_chunk = in_df.cix[output_index_to_input[idx], 0]
                        index_value, shape = (
                            src_chunk.index_value,
                            (src_chunk.shape[0],),
                        )
                    kw.update(
                        dict(
                            name=out_df.name,
                            dtype=out_df.dtype,
                            index=(idx,),
                            index_value=index_value,
                            shape=shape,
                        )
                    )
                elif op.output_types[0] == OutputType.tensor:
                    kw.update(dict(index=(0,), shape=(np.nan,), dtype=out_df.dtype))
                else:
                    kw.update(dict(index=(), shape=(), dtype=out_df.dtype))
            agg_chunks.append(chunk_op.new_chunk([chk], **kw))

        new_op = op.copy()
        if op.output_types[0] == OutputType.dataframe:
            if axis == 0:
                nsplits = ((out_df.shape[0],), tuple(c.shape[1] for c in agg_chunks))
            else:
                nsplits = (tuple(c.shape[0] for c in agg_chunks), (out_df.shape[1],))
            return new_op.new_tileables(
                op.inputs,
                chunks=agg_chunks,
                nsplits=nsplits,
                dtypes=out_df.dtypes,
                shape=out_df.shape,
                index_value=out_df.index_value,
                columns_value=out_df.columns_value,
            )
        elif op.output_types[0] == OutputType.series:
            nsplits = (tuple(c.shape[0] for c in agg_chunks),)
            return new_op.new_tileables(
                op.inputs,
                chunks=agg_chunks,
                nsplits=nsplits,
                dtype=out_df.dtype,
                shape=out_df.shape,
                index_value=out_df.index_value,
                name=out_df.name,
            )
        elif op.output_types[0] == OutputType.tensor:  # unique
            return new_op.new_tileables(
                op.inputs,
                chunks=agg_chunks,
                dtype=out_df.dtype,
                shape=out_df.shape,
                nsplits=((np.nan,),),
            )
        else:  # scalar
            return new_op.new_tileables(
                op.inputs, chunks=agg_chunks, dtype=out_df.dtype, shape=(), nsplits=()
            )

    @classmethod
    def tile(cls, op: "DataFrameAggregate"):
        in_df = op.inputs[0]

        if len(in_df.chunks) == 1:
            return cls._tile_single_chunk(op)
        elif not _agg_size_as_series and in_df.ndim == 2 and op.raw_func == "size":
            return (yield from cls._tile_size(op))
        else:
            return cls._tile_tree(op)

    @classmethod
    def _wrap_df(cls, op, value, index=None):
        xdf = cudf if op.gpu else pd
        axis = op.axis
        ndim = op.inputs[0].ndim

        if ndim == 2:
            dtype = None
            if isinstance(value, (np.generic, int, float, complex)):
                value = xdf.DataFrame([value], columns=index)
            elif not isinstance(value, xdf.DataFrame):
                new_index = None if not op.gpu else getattr(value, "index", None)
                dtype = getattr(value, "dtype", None)
                if xdf is pd:
                    value = xdf.DataFrame(value, columns=index, index=new_index)
                else:  # pragma: no cover
                    value = xdf.DataFrame(value)
                    value.index = new_index
                    value.columns = index
            else:
                return value

            value = value.T if axis == 0 else value
            if (
                dtype == np.dtype("O")
                and getattr(op.outputs[0], "dtypes", None) is not None
            ):
                value = value.astype(op.outputs[0].dtypes)
            return value
        else:
            if isinstance(value, (np.generic, int, float, complex)):
                value = xdf.Series([value], index=index)
            elif isinstance(value, np.ndarray):
                # assert value.ndim == 0
                value = xdf.Series(value.tolist(), index=index)
            return value

    @staticmethod
    def _pack_inputs(agg_funcs: List[ReductionAggStep], in_data):
        pos = 0
        out_dict = dict()
        for step in agg_funcs:
            if step.custom_reduction is None:
                out_dict[step.output_key] = in_data[pos]
            else:
                out_dict[step.output_key] = tuple(
                    in_data[pos : pos + step.output_limit]
                )
            pos += step.output_limit
        return out_dict

    @classmethod
    def _do_predefined_agg(cls, op: "DataFrameAggregate", input_obj, func_name, kwds):
        if func_name == "size":
            return input_obj.agg(lambda x: x.size, axis=op.axis)
        elif func_name == "str_concat":
            ret = input_obj.agg(lambda x: x.str.cat(**kwds), axis=op.axis)
            if isinstance(ret, str):
                ret = pd.Series([ret])
            return ret
        else:
            if op.gpu:
                if kwds.pop("numeric_only", None):
                    raise NotImplementedError("numeric_only not implemented under cudf")
            if isinstance(input_obj, pd.Index):
                kwds.pop("skipna", None)
            return getattr(input_obj, func_name)(**kwds)

    @classmethod
    def _select_dtypes(cls, in_data, op: "DataFrameAggregate"):
        if in_data.ndim == 2:
            if op.numeric_only:
                in_data = in_data.select_dtypes([np.number, np.bool_])
            elif op.bool_only:
                in_data = in_data.select_dtypes([np.bool_])
        return in_data

    @classmethod
    def _execute_map(cls, ctx, op: "DataFrameAggregate"):
        in_data = ctx[op.inputs[0].key]
        axis_index = op.outputs[0].index[op.axis]
        in_data = cls._select_dtypes(in_data, op)

        # map according to map groups
        ret_map_dfs = dict()
        in_cols_set = set(in_data.columns) if in_data.ndim == 2 else None
        for input_key, output_key, cols, func in op.pre_funcs:
            if cols and in_cols_set == set(cols):
                cols = None

            src_df = in_data if cols is None else in_data[cols]
            if input_key == output_key:
                ret_map_dfs[output_key] = src_df
            else:
                ret_map_dfs[output_key] = func(src_df, gpu=op.is_gpu())

        agg_dfs = []
        for (
            input_key,
            _,
            map_func_name,
            _agg_func_name,
            custom_reduction,
            _output_key,
            _output_limit,
            kwds,
        ) in op.agg_funcs:
            input_obj = ret_map_dfs[input_key]
            if map_func_name == "custom_reduction":
                pre_result = custom_reduction.pre(input_obj)
                if not isinstance(pre_result, tuple):
                    pre_result = (pre_result,)

                if custom_reduction.pre_with_agg:
                    # when custom_reduction.pre already aggregates, skip
                    agg_result = pre_result
                else:
                    agg_result = custom_reduction.agg(*pre_result)
                    if not isinstance(agg_result, tuple):
                        agg_result = (agg_result,)

                agg_dfs.extend(
                    [cls._wrap_df(op, r, index=[axis_index]) for r in agg_result]
                )
            else:
                agg_dfs.append(
                    cls._wrap_df(
                        op,
                        cls._do_predefined_agg(op, input_obj, map_func_name, kwds),
                        index=[axis_index],
                    )
                )
        ctx[op.outputs[0].key] = tuple(agg_dfs)

    @classmethod
    def _execute_combine(cls, ctx, op: "DataFrameAggregate"):
        in_data = ctx[op.inputs[0].key]
        in_data_dict = cls._pack_inputs(op.agg_funcs, in_data)
        axis = op.axis
        axis_index = op.outputs[0].index[axis]

        combines = []
        for (
            _input_key,
            _,
            _map_func_name,
            agg_func_name,
            custom_reduction,
            output_key,
            _output_limit,
            kwds,
        ) in op.agg_funcs:
            input_obj = in_data_dict[output_key]
            if agg_func_name == "custom_reduction":
                agg_result = custom_reduction.agg(*input_obj)
                if not isinstance(agg_result, tuple):
                    agg_result = (agg_result,)
                combines.extend(
                    [cls._wrap_df(op, r, index=[axis_index]) for r in agg_result]
                )
            else:
                combines.append(
                    cls._wrap_df(
                        op,
                        cls._do_predefined_agg(op, input_obj, agg_func_name, kwds),
                        index=[axis_index],
                    )
                )
        ctx[op.outputs[0].key] = tuple(combines)

    @classmethod
    def _execute_agg(cls, ctx, op: "DataFrameAggregate"):
        xdf = cudf if op.gpu else pd
        xp = cp if op.gpu else np

        out = op.outputs[0]
        in_data = ctx[op.inputs[0].key]
        in_data_dict = cls._pack_inputs(op.agg_funcs, in_data)
        axis = op.axis

        # perform agg
        for (
            _input_key,
            _,
            _map_func_name,
            agg_func_name,
            custom_reduction,
            output_key,
            _output_limit,
            kwds,
        ) in op.agg_funcs:
            input_obj = in_data_dict[output_key]
            if agg_func_name == "custom_reduction":
                agg_result = custom_reduction.agg(*input_obj)
                if not isinstance(agg_result, tuple):
                    agg_result = (agg_result,)
                in_data_dict[output_key] = custom_reduction.post(*agg_result)
            else:
                in_data_dict[output_key] = cls._do_predefined_agg(
                    op, input_obj, agg_func_name, kwds
                )

        aggs = []
        # perform post op
        for input_keys, _output_key, func_name, cols, func in op.post_funcs:
            if cols is None:
                func_inputs = [in_data_dict[k] for k in input_keys]
            else:
                func_inputs = [in_data_dict[k][cols] for k in input_keys]

            agg_series = func(*func_inputs, gpu=op.is_gpu())
            agg_series_ndim = getattr(agg_series, "ndim", 0)

            ser_index = None
            if agg_series_ndim < out.ndim:
                ser_index = [func_name]
            aggs.append(cls._wrap_df(op, agg_series, index=ser_index))

        # concatenate to produce final result
        concat_df = xdf.concat(aggs, axis=axis)
        if op.output_types[0] == OutputType.series:
            if concat_df.ndim > 1:
                if op.inputs[0].ndim == 2:
                    if axis == 0:
                        concat_df = concat_df.iloc[0, :]
                    else:
                        concat_df = concat_df.iloc[:, 0]
                else:
                    concat_df = concat_df.iloc[:, 0]
            concat_df.name = op.outputs[0].name

            concat_df = concat_df.astype(op.outputs[0].dtype, copy=False)
        elif op.output_types[0] == OutputType.scalar:
            concat_df = concat_df.iloc[0]
            try:
                concat_df = concat_df.astype(op.outputs[0].dtype)
            except AttributeError:
                # concat_df may be a string and has no `astype` method
                pass
        elif op.output_types[0] == OutputType.tensor:
            concat_df = xp.array(concat_df).astype(dtype=out.dtype)
        else:
            if axis == 0:
                concat_df = concat_df.reindex(op.outputs[0].index_value.to_pandas())
            else:
                concat_df = concat_df[op.outputs[0].columns_value.to_pandas()]

            concat_df = concat_df.astype(op.outputs[0].dtypes, copy=False)
        ctx[op.outputs[0].key] = concat_df

    @classmethod
    def _handle_cudf_agg_error(
        cls,
        func_name: Union[str, List[str], Callable, List[Callable]],
        exception: Exception,
    ):
        if callable(func_name) or (
            isinstance(func_name, list) and any([callable(f) for f in func_name])
        ):
            raise NotImplementedError(
                "Cudf does not support callable type aggregation functions for now. "
                "Please try to replace them with string type ones."
            ) from exception
        else:
            raise RuntimeError(f"Cudf error: {str(exception)}") from exception

    @classmethod
    def _cudf_agg(cls, op: "DataFrameAggregate", in_data):
        data = in_data
        if is_series(in_data):
            data = in_data.to_frame()
            # cudf to_frame doc: https://docs.rapids.ai/api/cudf/stable/api_docs/api/cudf.series.to_frame
            # The `name` option is None by default, but the actual behaviour sets the name to `0`
            # So define the columns manually
            data.columns = [in_data.name]

        func_name = (
            op.func_rename[0] or op.func[0] if len(op.func) == 1 else op.func_rename
        )
        # df.agg doesn't support the parameter axis in cudf
        # when axis is not None, cudf raises Error
        # ref: https://docs.rapids.ai/api/cudf/stable/api_docs/api/cudf.dataframe.agg
        axis = op.axis if op.axis == 1 else None
        result = None
        try:
            result = data.agg(func_name, axis=axis)
        except Exception as error:
            if isinstance(func_name, str):
                try:
                    result = getattr(data, func_name)(axis=axis)
                except Exception as e:
                    cls._handle_cudf_agg_error(func_name, e)
            else:
                cls._handle_cudf_agg_error(func_name, error)

        if is_series(result) and op.output_types[0] == OutputType.scalar:
            result = result[0]
        elif is_dataframe(result) and op.output_types[0] == OutputType.series:
            result = result.iloc[:, 0]
        return result

    @classmethod
    @redirect_custom_log
    @enter_current_session
    def execute(cls, ctx, op: "DataFrameAggregate"):
        if op.stage == OperandStage.map:
            cls._execute_map(ctx, op)
        elif op.stage == OperandStage.combine:
            cls._execute_combine(ctx, op)
        elif op.stage == OperandStage.agg:
            cls._execute_agg(ctx, op)
        elif not _agg_size_as_series and op.raw_func == "size":
            xp = cp if op.gpu else np
            ctx[op.outputs[0].key] = xp.array(
                ctx[op.inputs[0].key].agg(op.raw_func, axis=op.axis)
            ).reshape(op.outputs[0].shape)
        else:
            xp = cp if op.gpu else np
            in_obj = op.inputs[0]
            in_data = ctx[in_obj.key]
            in_data = cls._select_dtypes(in_data, op)
            if isinstance(in_obj, INDEX_CHUNK_TYPE):
                result = op.func[0](in_data)
            elif (
                op.output_types[0] == OutputType.scalar
                and in_data.shape == (0,)
                and callable(op.func[0])
            ):
                result = op.func[0](in_data)
            else:
                if is_cudf(in_data):
                    result = cls._cudf_agg(op, in_data)
                else:
                    result = (
                        in_data.agg(op.raw_func, axis=op.axis)
                        if op.raw_func is not None
                        else in_data.agg(**op.raw_func_kw, axis=op.axis)
                    )
                if op.outputs[0].ndim == 1:
                    result = result.astype(op.outputs[0].dtype, copy=False)

            if op.output_types[0] == OutputType.tensor:
                result = xp.array(result)
            ctx[op.outputs[0].key] = result


def is_funcs_aggregate(func, func_kw=None, ndim=2):
    func_kw = func_kw or dict()
    if ndim == 1 and func is None:
        func, func_kw = func_kw, dict()

    to_check = []
    if func is not None:
        if isinstance(func, (list, tuple)):
            to_check.extend(func)
        elif isinstance(func, dict):
            if ndim == 2:
                for f in func.values():
                    if isinstance(f, Iterable) and not isinstance(f, str):
                        to_check.extend(f)
                    else:
                        to_check.append(f)
            else:
                if any(isinstance(v, tuple) for v in func.values()):
                    raise TypeError("nested renamer is not supported")
                to_check.extend(func.values())
        else:
            to_check.append(func)
    else:
        for v in func_kw.values():
            if (
                not isinstance(v, tuple)
                or len(v) != 2
                or (not isinstance(v[1], str) and not callable(v[1]))
            ):
                raise TypeError("Must provide 'func' or tuples of (column, aggfunc).")
            else:
                to_check.append(v[1])

    compiler = ReductionCompiler(store_source=True)
    for f in to_check:
        if f in _agg_functions:
            continue
        elif callable(f):
            try:
                if ndim == 2:
                    compiler.add_function(f, 2, cols=["A", "B"])
                else:
                    compiler.add_function(f, 1)
            except ValueError:
                return False
        else:
            return False
    return True


def normalize_reduction_funcs(op, ndim=None):
    raw_func = op.raw_func
    if ndim == 1 and raw_func is None:
        raw_func = op.raw_func_kw

    if raw_func is not None:
        if isinstance(raw_func, dict):
            if ndim == 2:
                new_func = OrderedDict()
                for k, v in raw_func.items():
                    if isinstance(v, str) or callable(v):
                        new_func[k] = [v]
                    else:
                        new_func[k] = v
                op.func = new_func
            else:
                op.func = list(raw_func.values())
                op.func_rename = list(raw_func.keys())
        elif isinstance(raw_func, Iterable) and not isinstance(raw_func, str):
            op.func = list(raw_func)
        else:
            op.func = [raw_func]
    else:
        new_func = OrderedDict()
        new_func_names = OrderedDict()
        for k, v in op.raw_func_kw.items():
            try:
                col_funcs = new_func[v[0]]
                col_func_names = new_func_names[v[0]]
            except KeyError:
                col_funcs = new_func[v[0]] = []
                col_func_names = new_func_names[v[0]] = []
            col_funcs.append(v[1])
            col_func_names.append(k)
        op.func = new_func
        op.func_rename = functools.reduce(
            lambda a, b: a + b, new_func_names.values(), []
        )

    custom_idx = 0
    if isinstance(op.func, list):
        custom_iter = (f for f in op.func if isinstance(f, CustomReduction))
    else:
        custom_iter = (f for f in op.func.values() if isinstance(f, CustomReduction))
    for r in custom_iter:
        if r.name == "<custom>":
            r.name = f"<custom_{custom_idx}>"
            custom_idx += 1


def aggregate(df, func=None, axis=0, **kw):
    axis = validate_axis(axis, df)
    if (
        df.ndim == 2
        and isinstance(func, dict)
        and (df.op.output_types[0] == OutputType.series or axis == 1)
    ):
        raise NotImplementedError(
            "Currently cannot aggregate dicts over axis=1 on %s" % type(df).__name__
        )
    combine_size = kw.pop("_combine_size", None) or options.combine_size
    numeric_only = kw.pop("_numeric_only", None)
    bool_only = kw.pop("_bool_only", None)

    output_type = kw.pop("_output_type", None)
    dtypes = kw.pop("_dtypes", None)
    index = kw.pop("_index", None)
    raw_func_name = kw.pop("_func_name", None)

    if not is_funcs_aggregate(func, func_kw=kw, ndim=df.ndim):
        return df.transform(func, axis=axis, _call_agg=True)

    func_rename = (
        (
            [raw_func_name]
            if raw_func_name
            else (func if isinstance(func, list) else [func])
        )
        if df.op.gpu
        else None
    )

    op = DataFrameAggregate(
        raw_func=copy.deepcopy(func),
        raw_func_kw=copy.deepcopy(kw),
        func_rename=func_rename,
        axis=axis,
        combine_size=combine_size,
        numeric_only=numeric_only,
        bool_only=bool_only,
    )

    return op(df, output_type=output_type, dtypes=dtypes, index=index)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import configparser
import logging
import os
import time
from typing import Any, Dict, List, Optional, Tuple

from ....utils import get_default_logging_config_file_path
from ...constants import (
    DEFAULT_MARS_LOG_BACKUP_COUNT,
    DEFAULT_MARS_LOG_DIR,
    DEFAULT_MARS_LOG_FILE_NAME,
    DEFAULT_MARS_LOG_MAX_BYTES,
    MARS_LOG_DIR_KEY,
)
from ...oscar import create_actor_pool
from ...resource import Resource, cuda_count

logger = logging.getLogger(__name__)


def _need_suspend_sigint() -> bool:
    try:
        from IPython import get_ipython

        return get_ipython() is not None
    except ImportError:
        return False


def _get_root_logger_level_and_format() -> Tuple[str, Optional[str]]:
    root = logging.getLogger()
    level = logging.getLevelName(root.getEffectiveLevel())
    if level.startswith("WARN"):
        level = "WARN"
    handler = root.handlers[0] if root.handlers else None
    fmt = handler.formatter._fmt if handler else None
    return level, fmt


def _apply_log_level(level: Optional[str], config: configparser.RawConfigParser):
    if level is not None:
        loggers = config["loggers"]["keys"].split(",")
        for lg in loggers:
            lg = lg.strip()
            config[f"logger_{lg}"]["level"] = level.upper()
    return config


def _apply_log_format(fmt: Optional[str], config: configparser.RawConfigParser):
    if fmt is not None:
        formatters = config["formatters"]["keys"].split(",")
        for formatter in formatters:
            formatter = formatter.strip()
            config[f"formatter_{formatter}"]["format"] = fmt
    return config


def _apply_log_file_path(
    log_subdir: str,
    is_default_logging_config: bool,
    is_default_log_dir: bool,
    config: configparser.RawConfigParser,
):
    if is_default_logging_config:
        assert "handler_file_handler" in config
        log_file_path = os.path.join(log_subdir, DEFAULT_MARS_LOG_FILE_NAME)
        config["handler_file_handler"]["args"] = rf"(r'{log_file_path}',)"
        config["handler_file_handler"]["kwargs"] = (
            r"{'mode': 'a', "
            rf"'maxBytes': {DEFAULT_MARS_LOG_MAX_BYTES}, "
            rf"'backupCount': {DEFAULT_MARS_LOG_BACKUP_COUNT}"
            r"}"
        )
    elif not is_default_log_dir:
        # TODO: don't have a perfect way to handle this situation.
        raise ValueError(
            "Unable to change the log directory when using a user-defined logging"
            " configuration file."
        )
    return config


def _parse_file_logging_config(
    logging_conf: Dict[str, Any]
) -> Tuple[configparser.RawConfigParser, str]:
    """
    Parse the file logging config and apply logging configurations with higher priority.

    For interactive environments (from_cmd=False), the log level and format on the web follow our
    default configuration file, while the level and format on the console follow the user's
    configuration (logging.basicConfig) or keep the default.

    If env is cmd (from_cmd=True, e.g. user invokes `python -m mars.worker`),
    the log level and format on the web and console follow user's config.
    """
    from_cmd = logging_conf.get("from_cmd", False)
    log_level = logging_conf.get("level", None)
    log_fmt = logging_conf.get("format", None)
    # there's no need to add a prefix to the log subdir since it is shared by the supervisor and
    # workers in a local cluster.
    subdir_prefix = logging_conf.get("subdir_prefix", None) if from_cmd else None
    if MARS_LOG_DIR_KEY in os.environ:
        # in a local cluster, this env var may have been configured by another role.
        log_subdir = os.environ[MARS_LOG_DIR_KEY]
    else:
        log_subdir = _get_log_subdir(logging_conf.get("log_dir", None), subdir_prefix)
        # for FileLoggerActor(s).
        os.environ[MARS_LOG_DIR_KEY] = log_subdir
    log_config_path = _get_log_config_path(logging_conf.get("log_config", None))
    is_default_logging_config = logging_conf.get("log_config", None) is None
    is_default_log_dir = logging_conf.get("log_dir", None) is None

    config = configparser.RawConfigParser()
    config.read(log_config_path)
    _apply_log_format(log_fmt, config)
    _apply_log_level(log_level, config)
    _apply_log_file_path(
        log_subdir, is_default_logging_config, is_default_log_dir, config
    )

    # optimize logs for local runtimes (like IPython).
    if not from_cmd and is_default_logging_config:
        # console outputs follows user's configuration.
        root_level, root_fmt = _get_root_logger_level_and_format()

        assert "handler_stream_handler" in config
        config["handler_stream_handler"]["level"] = root_level or log_level or "WARN"
        if root_fmt:
            assert "formatter_console" not in config
            config.add_section("formatter_console")
            config["formatter_console"]["format"] = root_fmt
            config["formatters"]["keys"] += ",console"
            config["handler_stream_handler"]["formatter"] = "console"
    return config, log_config_path


def _get_or_create_default_log_dir() -> str:
    os.makedirs(DEFAULT_MARS_LOG_DIR, exist_ok=True)
    return DEFAULT_MARS_LOG_DIR


def _get_log_subdir(log_dir: Optional[str], subdir_prefix: Optional[str]) -> str:
    if log_dir is None:
        log_dir = _get_or_create_default_log_dir()
    if not os.path.exists(log_dir):
        raise RuntimeError(f"Log directory does not exist: {log_dir}")

    subdir_prefix = "" if subdir_prefix is None else subdir_prefix
    subdir_name = subdir_prefix + str(time.time_ns())
    log_subdir = os.path.join(log_dir, subdir_name)
    os.makedirs(log_subdir, exist_ok=True)
    return log_subdir


def _get_log_config_path(log_config: Optional[str]) -> str:
    if log_config is None:
        log_config = str(get_default_logging_config_file_path())
    if not os.path.exists(log_config):
        raise RuntimeError(f"Logging configuration file does not exist: {log_config}")
    return log_config


def _config_logging(
    logging_conf: Dict[str, Any]
) -> Optional[configparser.RawConfigParser]:
    parsed_logging_conf, log_config_path = _parse_file_logging_config(logging_conf)

    logging.config.fileConfig(
        parsed_logging_conf,
        disable_existing_loggers=False,
    )
    logger.info("Logging configurations from %s have been applied", log_config_path)
    logger.info("Logging directory %s", os.environ[MARS_LOG_DIR_KEY])

    return parsed_logging_conf


async def create_supervisor_actor_pool(
    address: str,
    n_process: int,
    modules: List[str] = None,
    ports: List[int] = None,
    subprocess_start_method: str = None,
    oscar_config: dict = None,
    **kwargs,
):
    logging_conf = kwargs.get("logging_conf", None)
    if logging_conf is None:
        logging_conf = dict()
    logging_conf["subdir_prefix"] = "supervisor_"
    logging_conf = _config_logging(logging_conf)
    kwargs["logging_conf"] = logging_conf
    if oscar_config:
        numa_config = oscar_config.get("numa", dict())
        numa_external_address_scheme = numa_config.get("external_addr_scheme", None)
        numa_enable_internal_address = numa_config.get("enable_internal_addr", True)
        external_address_schemes = [numa_external_address_scheme] * (n_process + 1)
        enable_internal_addresses = [numa_enable_internal_address] * (n_process + 1)
        extra_conf = oscar_config["extra_conf"]
    else:
        external_address_schemes = enable_internal_addresses = extra_conf = None
    return await create_actor_pool(
        address,
        n_process=n_process,
        ports=ports,
        external_address_schemes=external_address_schemes,
        enable_internal_addresses=enable_internal_addresses,
        modules=modules,
        subprocess_start_method=subprocess_start_method,
        suspend_sigint=_need_suspend_sigint(),
        extra_conf=extra_conf,
        **kwargs,
    )


async def create_worker_actor_pool(
    address: str,
    band_to_resource: Dict[str, Resource],
    n_io_process: int = 1,
    modules: List[str] = None,
    ports: List[int] = None,
    cuda_devices: List[int] = None,
    subprocess_start_method: str = None,
    oscar_config: dict = None,
    **kwargs,
):
    logging_conf = kwargs.get("logging_conf", None)
    if logging_conf is None:
        logging_conf = dict()
    logging_conf["subdir_prefix"] = "worker_"
    logging_conf = _config_logging(logging_conf)
    kwargs["logging_conf"] = logging_conf
    # TODO: support NUMA when ready
    n_process = sum(
        int(resource.num_cpus) or int(resource.num_gpus)
        for resource in band_to_resource.values()
    )
    envs = []
    labels = ["main"]

    oscar_config = oscar_config or dict()
    numa_config = oscar_config.get("numa", dict())
    numa_external_address_scheme = numa_config.get("external_addr_scheme")
    numa_enable_internal_address = numa_config.get("enable_internal_addr")
    gpu_config = oscar_config.get("gpu", dict())
    gpu_external_address_scheme = gpu_config.get("external_addr_scheme")
    gpu_enable_internal_address = gpu_config.get("enable_internal_addr")
    # get io process config
    io_config = oscar_config.get("io", dict())
    io_external_address_scheme = io_config.get("external_addr_scheme")
    io_enable_internal_address = io_config.get("enable_internal_addr")
    extra_conf = oscar_config.get("extra_conf", dict())

    if cuda_devices is None:  # pragma: no cover
        env_devices = os.environ.get("CUDA_VISIBLE_DEVICES")
        if not env_devices:
            cuda_devices = list(range(cuda_count()))
        else:
            cuda_devices = [int(i) for i in env_devices.split(",")]

    external_address_schemes = [numa_external_address_scheme]
    enable_internal_addresses = [numa_enable_internal_address]
    i_gpu = iter(sorted(cuda_devices))
    for band, resource in band_to_resource.items():
        if band.startswith("gpu"):
            idx = str(next(i_gpu))
            envs.append({"CUDA_VISIBLE_DEVICES": idx})
            labels.append(f"gpu-{idx}")
            external_address_schemes.append(gpu_external_address_scheme)
            enable_internal_addresses.append(gpu_enable_internal_address)
        else:
            assert band.startswith("numa")
            num_cpus = int(resource.num_cpus)
            if cuda_devices:
                # if has cuda device, disable all cuda devices for numa processes
                envs.extend([{"CUDA_VISIBLE_DEVICES": "-1"} for _ in range(num_cpus)])
            labels.extend([band] * num_cpus)
            external_address_schemes.extend(
                [numa_external_address_scheme for _ in range(num_cpus)]
            )
            enable_internal_addresses.extend(
                [numa_enable_internal_address for _ in range(num_cpus)]
            )

    # handle n_io_process
    n_process += n_io_process
    # sub-pools for IO(transfer and spill)
    for _ in range(n_io_process):
        if envs:  # pragma: no cover
            envs.append({"CUDA_VISIBLE_DEVICES": "-1"})
        labels.append("io")
        external_address_schemes.append(io_external_address_scheme)
        enable_internal_addresses.append(io_enable_internal_address)

    return await create_actor_pool(
        address,
        n_process=n_process,
        ports=ports,
        labels=labels,
        envs=envs,
        modules=modules,
        subprocess_start_method=subprocess_start_method,
        suspend_sigint=_need_suspend_sigint(),
        external_address_schemes=external_address_schemes,
        enable_internal_addresses=enable_internal_addresses,
        extra_conf=extra_conf,
        **kwargs,
    )



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...serialization.serializables import FieldTypes, KeyField, ListField
from ..array_utils import as_same_device, device
from ..core import TensorOrder
from ..datasource import tensor as astensor
from ..operands import TensorHasInput, TensorOperandMixin
from ..utils import reverse_order


def _reorder(x, axes):
    if x is None:
        return
    return type(x)(x[ax] for ax in axes)


class TensorTranspose(TensorHasInput, TensorOperandMixin):
    _op_type_ = OperandDef.TRANSPOSE

    _input = KeyField("input")
    _axes = ListField("axes", FieldTypes.int32)

    def __init__(self, axes=None, **kw):
        super().__init__(
            _axes=axes,
            # transpose will create a view
            create_view=True,
            **kw
        )

    @property
    def axes(self):
        return getattr(self, "_axes", None)

    def __call__(self, a):
        shape = tuple(
            s if np.isnan(s) else int(s) for s in _reorder(a.shape, self._axes)
        )
        if self._axes == list(reversed(range(a.ndim))):
            # order reversed
            tensor_order = reverse_order(a.order)
        else:
            tensor_order = TensorOrder.C_ORDER
        return self.new_tensor([a], shape, order=tensor_order)

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    def on_output_modify(self, new_output):
        op = self.copy().reset_key()
        return op(new_output)

    def on_input_modify(self, new_input):
        op = self.copy().reset_key()
        return op(new_input)

    @classmethod
    def tile(cls, op):
        tensor = op.outputs[0]

        out_chunks = []
        for c in op.inputs[0].chunks:
            chunk_op = op.copy().reset_key()
            chunk_shape = tuple(
                s if np.isnan(s) else int(s) for s in _reorder(c.shape, op.axes)
            )
            chunk_idx = _reorder(c.index, op.axes)
            out_chunk = chunk_op.new_chunk(
                [c], shape=chunk_shape, index=chunk_idx, order=tensor.order
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        nsplits = _reorder(op.inputs[0].nsplits, op.axes)
        return new_op.new_tensors(
            op.inputs,
            op.outputs[0].shape,
            order=tensor.order,
            chunks=out_chunks,
            nsplits=nsplits,
        )

    @classmethod
    def execute(cls, ctx, op):
        (x,), device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True
        )

        axes = op.axes
        with device(device_id):
            ctx[op.outputs[0].key] = xp.transpose(x, axes or None)


def transpose(a, axes=None):
    """
    Permute the dimensions of a tensor.

    Parameters
    ----------
    a : array_like
        Input tensor.
    axes : list of ints, optional
        By default, reverse the dimensions, otherwise permute the axes
        according to the values given.

    Returns
    -------
    p : Tensor
        `a` with its axes permuted.  A view is returned whenever
        possible.

    See Also
    --------
    moveaxis
    argsort

    Notes
    -----
    Use `transpose(a, argsort(axes))` to invert the transposition of tensors
    when using the `axes` keyword argument.

    Transposing a 1-D array returns an unchanged view of the original tensor.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> x = mt.arange(4).reshape((2,2))
    >>> x.execute()
    array([[0, 1],
           [2, 3]])

    >>> mt.transpose(x).execute()
    array([[0, 2],
           [1, 3]])

    >>> x = mt.ones((1, 2, 3))
    >>> mt.transpose(x, (1, 0, 2)).shape
    (2, 1, 3)

    """
    a = astensor(a)
    if axes:
        if len(axes) != a.ndim:
            raise ValueError("axes don't match tensor")

    if not axes:
        axes = list(range(a.ndim))[::-1]
    else:
        axes = list(axes)
    op = TensorTranspose(axes, dtype=a.dtype, sparse=a.issparse())
    return op(a)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .all import TensorAll, all
from .allclose import allclose
from .any import TensorAny, any
from .argmax import TensorArgmax, argmax
from .argmin import TensorArgmin, argmin
from .array_equal import array_equal
from .count_nonzero import TensorCountNonzero, count_nonzero
from .cumprod import TensorCumprod, cumprod
from .cumsum import TensorCumsum, cumsum
from .max import TensorMax, max
from .mean import TensorMean, mean
from .min import TensorMin, min
from .nanargmax import TensorNanArgmax, nanargmax
from .nanargmin import TensorNanArgmin, nanargmin
from .nancumprod import TensorNanCumprod, nancumprod
from .nancumsum import TensorNanCumsum, nancumsum
from .nanmax import TensorNanMax, nanmax
from .nanmean import TensorNanMean, nanmean
from .nanmin import TensorNanMin, nanmin
from .nanprod import TensorNanProd, nanprod
from .nanstd import nanstd
from .nansum import TensorNanSum, nansum
from .nanvar import TensorNanMoment, TensorNanVar, nanvar
from .prod import TensorProd, prod
from .std import std
from .sum import TensorSum, sum
from .var import TensorMoment, TensorVar, var


def _install():
    from ..core import Tensor, TensorData

    for cls in (Tensor, TensorData):
        setattr(cls, "sum", sum)
        setattr(cls, "prod", prod)
        setattr(cls, "max", max)
        setattr(cls, "min", min)
        setattr(cls, "all", all)
        setattr(cls, "any", any)
        setattr(cls, "mean", mean)
        setattr(cls, "argmax", argmax)
        setattr(cls, "argmin", argmin)
        setattr(cls, "cumsum", cumsum)
        setattr(cls, "cumprod", cumprod)
        setattr(cls, "var", var)
        setattr(cls, "std", std)


_install()
del _install



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import List, Tuple, Union

import numpy as np

from ...core import ChunkType
from ...tensor import arithmetic, reduction
from ...tensor.fuse import TensorNeFuseChunk
from ...tensor.fuse.numexpr import NUMEXPR_INSTALLED
from .core import REDUCTION, GraphTraversalOptimizer, _Fuse, register_optimizer

logger = logging.getLogger(__name__)


REDUCTION_OP = {
    reduction.TensorSum,
    reduction.TensorProd,
    reduction.TensorMax,
    reduction.TensorMin,
}
SUPPORT_OP = {
    arithmetic.TensorAdd,
    arithmetic.TensorSubtract,
    arithmetic.TensorMultiply,
    arithmetic.TensorDivide,
    arithmetic.TensorPower,
    arithmetic.TensorMod,
    arithmetic.TensorNegative,
    arithmetic.TensorAbs,
    arithmetic.TensorConj,
    arithmetic.TensorExp,
    arithmetic.TensorLog,
    arithmetic.TensorLog10,
    arithmetic.TensorExpm1,
    arithmetic.TensorLog1p,
    arithmetic.TensorSqrt,
    arithmetic.TensorEqual,
    arithmetic.TensorNotEqual,
    arithmetic.TensorLessThan,
    arithmetic.TensorLessEqual,
    arithmetic.TensorGreaterThan,
    arithmetic.TensorGreaterEqual,
    arithmetic.TensorSin,
    arithmetic.TensorCos,
    arithmetic.TensorTan,
    arithmetic.TensorArcsin,
    arithmetic.TensorArccos,
    arithmetic.TensorArctan,
    arithmetic.TensorSinh,
    arithmetic.TensorCosh,
    arithmetic.TensorTanh,
    arithmetic.TensorArcsinh,
    arithmetic.TensorArccosh,
    arithmetic.TensorArctanh,
    arithmetic.TensorLshift,
    arithmetic.TensorRshift,
    arithmetic.TensorTreeAdd,
    arithmetic.TensorTreeMultiply,
    arithmetic.TensorFloor,
    arithmetic.TensorCeil,
    arithmetic.TensorAnd,
    arithmetic.TensorOr,
    arithmetic.TensorNot,
    reduction.TensorSum,
    reduction.TensorProd,
    reduction.TensorMax,
    reduction.TensorMin,
}


@register_optimizer
class NumexprRuntimeOptimizer(GraphTraversalOptimizer):
    engine = "numexpr"

    def _can_fuse(self, node: ChunkType) -> Union[bool, object]:
        op = node.op
        op_type = type(op)
        if op_type in REDUCTION_OP:
            if len(op.axis) == 1 or len(op.axis) == node.ndim:
                return REDUCTION
            else:
                return False
        # return op_type in SUPPORT_OP
        if op_type not in SUPPORT_OP:
            return False
        if op_type in (arithmetic.TensorOr, arithmetic.TensorAnd):
            # numexpr only support logical and or:
            # https://numexpr.readthedocs.io/projects/NumExpr3/en/latest/user_guide.html#supported-operators
            if np.isscalar(op.lhs) or np.isscalar(op.rhs):  # pragma: no cover
                return False
        return True

    @classmethod
    def is_available(cls) -> bool:
        return NUMEXPR_INSTALLED

    def optimize(self) -> Tuple[List[_Fuse], List[ChunkType]]:
        fuses = self._graph_traverse(
            logger, "Refused fusing for numexpr because the tail node count > 1."
        )

        if fuses == ([], []):
            return [], []

        return self._fuse_nodes(fuses, TensorNeFuseChunk)



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
import json
import os
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor, as_completed
from queue import Queue
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import cloudpickle
import numpy as np
import pyarrow as pa
import pyarrow.compute as pc

from ...._mars.core.entity import OutputType
from ...._mars.serialization.serializables import DictField, Int32Field, StringField
from ...._mars.utils import lazy_import
from ...operand import DataOperand, DataOperandMixin

if TYPE_CHECKING:
    from fsspec import AbstractFileSystem
else:
    AbstractFileSystem = lazy_import("fsspec.AbstractFileSystem")


class HuggingfaceExport(DataOperand, DataOperandMixin):
    path: str = StringField("path")
    storage_options: Dict = DictField("storage_options")
    max_chunk_rows: int = Int32Field("max_chunk_rows")
    column_groups: dict = DictField("column_groups")
    num_threads: int = Int32Field("num_threads")
    version: str = StringField("version")

    def __call__(self, dataset):
        return self.new_tileable([dataset], **dataset.params)

    @classmethod
    def tile(cls, op: "HuggingfaceExport"):
        assert len(op.inputs) == 1
        inp = op.inputs[0]
        out = op.outputs[0]
        out_chunks = []
        for chunk in inp.chunks:
            chunk_op = op.copy().reset_key()
            out_chunk = chunk_op.new_chunk([chunk], index=chunk.index)
            out_chunks.append(out_chunk)
        return op.copy().new_tileable(
            op.inputs,
            chunks=out_chunks,
            nsplits=((np.nan,) * len(out_chunks), (np.nan,)),
            **out.params,
        )

    @classmethod
    def execute(cls, ctx, op: "HuggingfaceExport"):
        inp = ctx[op.inputs[0].key]
        out = op.outputs[0]
        r = ArrowWriter.ensure_path(op.path, op.storage_options).write(
            dataset=inp,
            chunk_index=out.index[0],
            max_chunk_rows=op.max_chunk_rows,
            column_groups=op.column_groups,
            num_threads=op.num_threads,
            version=op.version,
        )
        ctx[out.key] = r


@dataclasses.dataclass
class SchemaInfo:
    schema: pa.Schema
    column_groups: Dict
    max_chunk_rows: int


class ArrowWriter:
    _WRITER_CLASS = pa.RecordBatchStreamWriter
    _DEFAULT_VERSION = "0.0.0"
    _DEFAULT_MAX_CHUNK_ROWS = 100
    _META_DIR = ".meta"
    _MULTIMEDIA_PREFIX = "mdata"
    _DATA_PREFIX = "data"
    _FILE_NAME_FORMATTER = "{chunk_index}_{index}.arrow"
    _META_SCHEMA_INFO_KEY = b"xdataset_schema_info"

    def __init__(
        self,
        fs: AbstractFileSystem,
        path: Union[str, os.PathLike],
    ):
        self._fs = fs
        self._path = path

    @classmethod
    def ensure_path(
        cls,
        path: Union[str, os.PathLike],
        storage_options: Optional[dict] = None,
        create_if_not_exists: Optional[bool] = True,
    ):
        import fsspec

        fs_token_paths = fsspec.get_fs_token_paths(
            path, storage_options=storage_options
        )
        fs: fsspec.AbstractFileSystem = fs_token_paths[0]
        path = fs_token_paths[2][0]
        if not fs.exists(path):
            if create_if_not_exists:
                fs.mkdirs(path, exist_ok=True)
            else:
                raise Exception(f"The path {path} does not exist.")
        elif not fs.isdir(path):
            raise Exception(f"The path {path} should be a dir.")
        return cls(fs, path)

    def check_version(
        self, version: Optional[str] = None, overwrite: Optional[bool] = True
    ):
        version = version or self._DEFAULT_VERSION
        version_path = os.path.join(self._path, version)
        if self._fs.exists(version_path):
            if overwrite:
                self._fs.rm(version_path, recursive=True)
            else:
                raise Exception(f"The version {version_path} already exists.")

    def _embed_and_write_table(self, file: str, pa_table: pa.Table, features, info):
        """Write a Table to file.

        Args:
            example: the Table to add.
        """
        from datasets.features.features import require_storage_embed
        from datasets.table import embed_array_storage

        with self._fs.open(file, "wb") as stream:
            # TODO(codingl2k1): embed uid to table to check the different parts
            # of data are matched.
            schema = pa_table.schema.remove_metadata()
            with self._WRITER_CLASS(stream, schema) as writer:
                arrays = [
                    embed_array_storage(pa_table[name], feature)
                    if require_storage_embed(feature)
                    else pa_table[name]
                    for name, feature in features.items()
                ]
                pa_table = pa.Table.from_arrays(arrays, schema=schema)
                # Update embedded table nbytes to num_bytes
                info["num_bytes"] = pa_table.nbytes
                writer.write(pa_table)

    def _write_group_meta(
        self,
        q: Queue,
        meta_path: str,
        meta_table: pa.Table,
        name: str,
        schema_info: SchemaInfo,
    ):
        self._fs.mkdirs(meta_path, exist_ok=True)
        meta_table = meta_table.select([name]).flatten()
        group_schema_table = schema_info.schema.empty_table().select(
            schema_info.column_groups[name]
        )
        info = {
            "num_bytes": pc.sum(meta_table[f"{name}.num_bytes"]).as_py(),
            "num_rows": pc.sum(meta_table[f"{name}.num_rows"]).as_py(),
            "num_columns": group_schema_table.num_columns,
            "num_files": meta_table.num_rows,
            "schema_string": group_schema_table.schema.to_string(
                show_field_metadata=False, show_schema_metadata=False
            ),
        }
        q.put(info)
        info_file = os.path.join(meta_path, "info.json")
        with self._fs.open(info_file, "w") as f:
            json.dump(info, f)
        # Write index arrow table.
        index_file = os.path.join(meta_path, "index.arrow")
        with self._fs.open(index_file, "wb") as stream:
            with self._WRITER_CLASS(stream, meta_table.schema) as writer:
                writer.write(meta_table)
        # Write schema empty table.
        schema_file = os.path.join(meta_path, "schema.arrow")
        with self._fs.open(schema_file, "wb") as stream:
            with self._WRITER_CLASS(stream, group_schema_table.schema) as writer:
                writer.write(group_schema_table)
        return info

    def write_meta(
        self,
        meta_table: pa.Table,
        num_threads: Optional[int] = None,
        version: Optional[str] = None,
    ):
        version_path = os.path.join(self._path, version or self._DEFAULT_VERSION)
        schema_info = cloudpickle.loads(
            meta_table.schema.metadata[self._META_SCHEMA_INFO_KEY]
        )

        futures = []
        with ThreadPoolExecutor(
            max_workers=num_threads, thread_name_prefix=self.write_meta.__qualname__
        ) as executor:
            q: Queue = Queue()

            for name in meta_table.column_names:
                meta_path = os.path.join(version_path, name, self._META_DIR)
                futures.append(
                    executor.submit(
                        self._write_group_meta,
                        q,
                        meta_path,
                        meta_table,
                        name,
                        schema_info,
                    )
                )

            group_info = q.get()
            info = {
                "groups": meta_table.column_names,
                "num_rows": group_info["num_rows"],
                "max_chunk_rows": schema_info.max_chunk_rows,
            }
            info_file = os.path.join(version_path, "info.json")
            with self._fs.open(info_file, "w") as f:
                json.dump(info, f)

        # Raise exception if exists.
        return dict(zip(meta_table.column_names, (fut.result() for fut in futures)))

    def write(
        self,
        dataset,
        chunk_index: int,
        max_chunk_rows: Optional[int] = None,
        column_groups: Optional[dict] = None,
        num_threads: Optional[int] = None,
        version: Optional[str] = None,
    ):
        from datasets import Dataset
        from datasets.features.audio import Audio
        from datasets.features.image import Image

        assert isinstance(dataset, Dataset)

        if column_groups is None:
            # Auto split columns to multimedia and data groups.
            # TODO(codingl2k1): Should we always split dataset?
            multimedia_columns = []
            data_columns = []
            for idx, ft in enumerate(dataset.features.values()):
                if isinstance(ft, (Image, Audio)):
                    multimedia_columns.append(idx)
                else:
                    data_columns.append(idx)
            column_groups = {}
            if multimedia_columns:
                column_groups[self._MULTIMEDIA_PREFIX] = multimedia_columns
            if data_columns:
                column_groups[self._DATA_PREFIX] = data_columns

        feature_groups = []
        data_table = dataset.data
        for name, columns in column_groups.items():
            features = {}
            for col in columns:
                col_name = data_table.field(col).name
                features[col_name] = dataset.features[col_name]
            feature_groups.append(features)

        version_path = os.path.join(self._path, version or self._DEFAULT_VERSION)
        for name in column_groups.keys():
            self._fs.mkdirs(os.path.join(version_path, name), exist_ok=True)

        schema_info = None
        meta_info: Dict[str, List[Any]] = defaultdict(list)
        futures = []
        with ThreadPoolExecutor(
            max_workers=num_threads, thread_name_prefix=self.write.__qualname__
        ) as executor:
            # TODO(codingl2k1): split by max_chunk_rows may lead to rank data skew.
            # e.g. if 1000, 1000, 80, then the rank 2 world 3 worker read very few data.
            max_chunk_rows = max_chunk_rows or self._DEFAULT_MAX_CHUNK_ROWS
            for idx, pa_table in enumerate(
                dataset.with_format("arrow").iter(max_chunk_rows)
            ):
                if chunk_index == 0 and idx == 0:
                    schema_info = SchemaInfo(
                        pa_table.schema, column_groups, max_chunk_rows
                    )
                pa_table = pa_table.combine_chunks()
                for (name, columns), features in zip(
                    column_groups.items(), feature_groups
                ):
                    pa_group_table = pa_table.select(columns)
                    # The schema is not changed, don't need to table_cast.
                    filename = self._FILE_NAME_FORMATTER.format(
                        chunk_index=chunk_index, index=idx
                    )
                    file = os.path.join(version_path, name, filename)
                    file_info = {
                        "index": (chunk_index, idx),
                        "num_rows": pa_table.num_rows,
                        "num_bytes": pa_group_table.nbytes,
                    }
                    meta_info[name].append(file_info)
                    futures.append(
                        executor.submit(
                            self._embed_and_write_table,
                            file,
                            pa_group_table,
                            features,
                            file_info,
                        )
                    )
            # Raise exception if exists.
            for fut in as_completed(futures):
                fut.result()

        struct_schema = pa.struct(
            [
                ("index", pa.list_(pa.int32(), 2)),
                ("num_rows", pa.int64()),
                ("num_bytes", pa.int64()),
            ]
        )
        meta_table = pa.Table.from_pydict(
            meta_info,
            schema=pa.schema(
                [pa.field(k, struct_schema) for k in meta_info],
                metadata=None
                if schema_info is None
                else {self._META_SCHEMA_INFO_KEY: cloudpickle.dumps(schema_info)},
            ),
        )
        return meta_table


def export(
    dataset,
    path: Union[str, os.PathLike],
    storage_options: Optional[dict] = None,
    create_if_not_exists: Optional[bool] = True,
    max_chunk_rows: Optional[int] = None,
    column_groups: Optional[dict] = None,
    num_threads: Optional[int] = None,
    version: Optional[str] = None,
    overwrite: Optional[bool] = True,
):
    arrow_writer = ArrowWriter.ensure_path(path, storage_options, create_if_not_exists)
    arrow_writer.check_version(version, overwrite)
    op = HuggingfaceExport(
        output_types=[OutputType.object],
        path=path,
        storage_options=storage_options,
        max_chunk_rows=max_chunk_rows,
        column_groups=column_groups,
        num_threads=num_threads,
        version=version,
    )
    meta_table = op(dataset).execute().fetch()
    info = arrow_writer.write_meta(meta_table)
    return info



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import re
from itertools import tee
from typing import List, Set, Text, Tuple

import numpy as np
import pandas as pd
from scipy.integrate import quad as integrate

NON_ALPHA = re.compile("\\W", re.UNICODE)
MAX_HASH = np.uint64((1 << 32) - 1)
MERSENNE_PRIME = np.uint64((1 << 61) - 1)


def ngrams(sequence: List[Text], n: int, min_length: int = 5):
    """
    Return the ngrams generated from a sequence of items, as an iterator.

    This is copied from https://github.com/ChenghaoMou/text-dedup.

    Parameters
    ----------
    sequence : List[Text]
        The sequence of items.
    n : int
        The length of each ngram.
    min_length : int, optional
        The minimum length of each ngram, by default 5

    Returns
    -------
    iterator
        The ngrams.

    Examples
    --------
    >>> list(ngrams(["a", "b", "c", "d"], 2, min_length=1))
    [('a', 'b'), ('b', 'c'), ('c', 'd')]
    >>> list(ngrams(["a", "b", "c", "d"], 2, min_length=5))
    []
    >>> list(ngrams(["a", "b"], 3, min_length=1))
    [('a', 'b')]
    """
    if len(sequence) < min_length:  # pragma: no cover
        return []
    if len(sequence) < n:  # pragma: no cover
        return [tuple(sequence)]
    iterables = tee(iter(sequence), n)
    for i, sub_iterable in enumerate(iterables):
        for _ in range(i):
            next(sub_iterable, None)
    return zip(*iterables)


def sha1_hash(data: bytes, d: int = 32) -> int:
    """
    Generate a d-bit hash value from the given data.

    This is copied from https://github.com/ChenghaoMou/text-dedup.

    Parameters
    ----------
    data : bytes
        The data to be hashed.
    d : int
        The number of bits of the hash value.

    Returns
    -------
    int
        The hash value.

    Examples
    --------
    >>> sha1_hash(b"hello world", 32)
    896314922
    >>> sha1_hash(b"hello world", 64)
    13028719972609469994
    >>> sha1_hash(b"hello world", 128)
    310522945683037930239412421226792791594
    """
    return int.from_bytes(hashlib.sha1(data).digest()[: d // 8], byteorder="little")


def optimal_param(
    threshold: float,
    num_perm: int,
    false_positive_weight: float = 0.5,
    false_negative_weight: float = 0.5,
) -> Tuple[int, int]:
    """
    Compute the optimal `MinHashLSH` parameter that minimizes the weighted sum
    of probabilities of false positive and false negative, taken from datasketch.

    You can also refer to the interactive demo at https://huggingface.co/spaces/bigcode/near-deduplication.
    This is copied from https://github.com/ChenghaoMou/text-dedup.

    Parameters
    ----------
    threshold : float
        The threshold for similarity.
    num_perm : int
        The number of permutations.
    false_positive_weight : float
        The weight of false positive.
    false_negative_weight : float
        The weight of false negative.

    Returns
    -------
    Tuple[int, int]
        The optimal `b` (bands) and `r` (rows) parameters.

    Examples
    --------
    >>> optimal_param(0.75, 256)
    (21, 12)
    >>> optimal_param(0.75, 256, 0.1, 0.9)
    (28, 9)
    """

    def false_positive_area(threshold: float, b: int, r: int):
        """Source: `datasketch.lsh`"""

        def proba(s):
            return 1 - (1 - s ** float(r)) ** float(b)

        a, _ = integrate(proba, 0.0, threshold)
        return a

    def false_negative_area(threshold: float, b: int, r: int):
        """Source: `datasketch.lsh`"""

        def proba(s):
            return 1 - (1 - (1 - s ** float(r)) ** float(b))

        a, _ = integrate(proba, threshold, 1.0)
        return a

    min_error = float("inf")
    opt = (0, 0)
    for b in range(1, num_perm + 1):
        max_r = int(num_perm / b)
        for r in range(1, max_r + 1):
            fp = false_positive_area(threshold, b, r)
            fn = false_negative_area(threshold, b, r)
            error = fp * false_positive_weight + fn * false_negative_weight
            if error < min_error:
                min_error = error
                opt = (b, r)
    return opt


def minhash_embed_func(
    row: pd.Series,
    *,
    text: str,
    num_perm: int,
    ngram_size: int,
    min_length: int,
    hashranges: List[Tuple[int, int]],
    permutations: np.ndarray,
) -> pd.Series:
    """
    Calculate hash values for the content.

    This is a modified version of https://github.com/ChenghaoMou/text-dedup.

    Parameters
    ----------
    row : pd.Series
        The row content to be embedded.
    text : str
        The text column of the columns.
    num_perm : int
        The number of permutations.
    ngram_size : int
        The size of n-grams.
    min_length : int
        The minimum length of the document in terms of tokens.
    hashranges : List[Tuple[int, int]]
        The ranges of hash values.
    permutations : np.ndarray
        The permutations for the minhash.

    Returns
    -------
    Dict[str, Any]
        The hash values in each range and the index.

    Examples
    --------
    >>> row = pd.Series({"text": "hello world", "__dedup_id": 0})
    >>> text = "text"
    >>> num_perm = 250
    >>> ngram_size = 1
    >>> hashranges = [(i, i + 25) for i in range(0, 250, 25)]
    >>> PERMUTATIONS = np.array(
    ...     [
    ...         (
    ...             RNG.randint(1, MERSENNE_PRIME, dtype=np.uint64),
    ...             RNG.randint(0, MERSENNE_PRIME, dtype=np.uint64),
    ...         )
    ...         for _ in range(num_perm)
    ...     ],
    ...     dtype=np.uint64,
    ... ).T
    >>> res = embed_func(content, idx, num_perm=num_perm, ngram_size=ngram_size, min_length=0, hashranges=hashranges, permutations=PERMUTATIONS)
    >>> len(res["__signatures"])
    10
    """
    content, idx = row[text], row["__dedup_id"]

    a, b = permutations

    masks: np.ndarray = np.full(shape=num_perm, dtype=np.uint64, fill_value=MAX_HASH)
    tokens: Set[str] = {
        " ".join(t) for t in ngrams(NON_ALPHA.split(content), ngram_size, min_length)
    }

    hashvalues: np.ndarray = np.array(
        [sha1_hash(token.lower().encode("utf-8")) for token in tokens], dtype=np.uint64
    )

    permuted_hashvalues = np.bitwise_and(
        ((hashvalues * np.tile(a, (len(hashvalues), 1)).T).T + b) % MERSENNE_PRIME,
        MAX_HASH,
    )
    hashvalues = np.vstack([permuted_hashvalues, masks]).min(axis=0)

    Hs = [
        (i, bytes(hashvalues[start:end].byteswap().data))
        for i, (start, end) in enumerate(hashranges)
    ]
    return pd.Series({"__signatures": Hs, "__id": idx})



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from typing import Any, List, Tuple, Type

import numpy as np

from ... import opcodes as OperandDef
from ...config import options
from ...core import TILEABLE_TYPE
from ...core.operand import OperandStage
from ...serialization.serializables import AnyField, Int32Field, Int64Field, StringField
from ...typing import ChunkType, TileableType
from ...utils import has_unknown_shape
from ..array_utils import as_same_device, device
from ..core import TENSOR_TYPE, TensorOrder
from ..datasource.array import tensor as astensor
from ..operands import TensorOperand, TensorOperandMixin


class TensorSearchsorted(TensorOperand, TensorOperandMixin):
    _op_type_ = OperandDef.SEARCHSORTED

    v = AnyField("v")
    side = StringField("side")
    combine_size = Int32Field("combine_size")
    # for chunk
    offset = Int64Field("offset")
    size = Int64Field("size")
    n_chunk = Int64Field("n_chunk")

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if isinstance(self.v, TILEABLE_TYPE):
            self.v = self._inputs[1]

    def __call__(self, a, v):
        inputs = [a]
        if isinstance(v, TILEABLE_TYPE):
            inputs.append(v)
            shape = v.shape
        else:
            shape = ()
        return self.new_tensor(inputs, shape=shape, order=TensorOrder.C_ORDER)

    @classmethod
    def _tile_one_chunk(cls, op, a, v, out):
        chunks = []
        if len(op.inputs) == 1:
            v_chunks = [v]
        else:
            v_chunks = v.chunks
        for v_chunk in v_chunks:
            chunk_op = op.copy().reset_key()
            in_chunks = [a.chunks[0]]
            if len(op.inputs) == 2:
                in_chunks.append(v_chunk)
            v_shape = v_chunk.shape if hasattr(v_chunk, "shape") else ()
            chunk_idx = v_chunk.index if len(op.inputs) == 2 else (0,)
            chunk = chunk_op.new_chunk(
                in_chunks, shape=v_shape, index=chunk_idx, order=out.order
            )
            chunks.append(chunk)
        new_op = op.copy().reset_key()
        nsplits = ((s,) for s in out.shape) if len(op.inputs) == 1 else v.nsplits
        return new_op.new_tensors(op.inputs, out.shape, chunks=chunks, nsplits=nsplits)

    @classmethod
    def _combine_chunks(
        cls,
        to_combine: List[ChunkType],
        op_type: Type,
        v: Any,
        stage: OperandStage,
        chunk_index: Tuple[int],
    ):
        from ..merge import TensorStack

        dtype = np.dtype(np.intp)
        v_shape = v.shape if hasattr(v, "shape") else ()
        combine_op = TensorStack(axis=0, dtype=dtype)
        combine_chunk = combine_op.new_chunk(to_combine, shape=v_shape)
        chunk_op = op_type(dtype=dtype, axis=(0,), stage=stage)
        return chunk_op.new_chunk(
            [combine_chunk], shape=v_shape, index=chunk_index, order=TensorOrder.C_ORDER
        )

    @classmethod
    def _tile_tree_reduction(
        cls, op: "TensorSearchsorted", a: TileableType, v: Any, out: TileableType
    ):
        from ..indexing import TensorSlice
        from ..merge import TensorConcatenate
        from ..reduction import TensorMax, TensorMin

        if has_unknown_shape(a):
            yield

        combine_size = op.combine_size or options.combine_size
        n_chunk = len(a.chunks)
        input_len = len(op.inputs)
        v_chunks = [v] if input_len == 1 else v.chunks
        cum_nsplits = [0] + np.cumsum(a.nsplits[0]).tolist()

        input_chunks = []
        offsets = []
        for i in range(n_chunk):
            offset = cum_nsplits[i]
            cur_chunk = a.chunks[i]
            chunk_size = a.shape[0]
            chunks = []
            if i > 0:
                last_chunk = a.chunks[i - 1]
                if last_chunk.shape[0] > 0:
                    slice_chunk_op = TensorSlice(
                        slices=[slice(-1, None)], dtype=cur_chunk.dtype
                    )
                    slice_chunk = slice_chunk_op.new_chunk(
                        [last_chunk], shape=(1,), order=out.order
                    )
                    chunks.append(slice_chunk)
                    chunk_size += 1
                    offset -= 1
            chunks.append(cur_chunk)
            if i < n_chunk - 1:
                next_chunk = a.chunks[i + 1]
                if next_chunk.shape[0] > 0:
                    slice_chunk_op = TensorSlice(
                        slices=[slice(1)], dtype=cur_chunk.dtype
                    )
                    slice_chunk = slice_chunk_op.new_chunk(
                        [next_chunk], shape=(1,), order=out.order
                    )
                    chunks.append(slice_chunk)
                    chunk_size += 1

            concat_op = TensorConcatenate(dtype=cur_chunk.dtype)
            concat_chunk = concat_op.new_chunk(
                chunks, shape=(chunk_size,), order=out.order, index=cur_chunk.index
            )
            input_chunks.append(concat_chunk)
            offsets.append(offset)

        out_chunks = []
        for v_chunk in v_chunks:
            chunks = []
            v_shape = v_chunk.shape if hasattr(v_chunk, "shape") else ()
            v_index = v_chunk.index if hasattr(v_chunk, "index") else (0,)
            for inp_chunk, offset in zip(input_chunks, offsets):
                chunk_op = op.copy().reset_key()
                chunk_op.stage = OperandStage.map
                chunk_op.offset = offset
                chunk_op.n_chunk = n_chunk
                chunk_op.size = a.shape[0]
                chunk_inputs = [inp_chunk]
                if input_len > 1:
                    chunk_inputs.append(v_chunk)
                map_chunk = chunk_op.new_chunk(
                    chunk_inputs, shape=v_shape, index=inp_chunk.index, order=out.order
                )
                chunks.append(map_chunk)

            op_type = TensorMax if op.side == "right" else TensorMin
            while len(chunks) > combine_size:
                new_chunks = []
                it = itertools.count(0)
                while True:
                    j = next(it)
                    to_combine = chunks[j * combine_size : (j + 1) * combine_size]
                    if len(to_combine) == 0:
                        break

                    new_chunks.append(
                        cls._combine_chunks(
                            to_combine, op_type, v_chunk, OperandStage.combine, (j,)
                        )
                    )
                chunks = new_chunks

            chunk = cls._combine_chunks(
                chunks, op_type, v_chunk, OperandStage.agg, v_index
            )
            out_chunks.append(chunk)

        new_op = op.copy().reset_key()
        nsplits = ((s,) for s in out.shape) if len(op.inputs) == 1 else v.nsplits
        return new_op.new_tensors(
            op.inputs, out.shape, chunks=out_chunks, nsplits=nsplits
        )

    @classmethod
    def tile(cls, op):
        a = op.inputs[0]
        out = op.outputs[0]
        input_len = len(op.inputs)
        if input_len == 1:
            v = op.v
        else:
            v = op.inputs[1]

        if len(a.chunks) == 1:
            return cls._tile_one_chunk(op, a, v, out)
        return (yield from cls._tile_tree_reduction(op, a, v, out))

    @classmethod
    def _execute_without_stage(cls, xp, a, v, op):
        return xp.searchsorted(a, v, side=op.side)

    @classmethod
    def _execute_map(cls, xp: Any, a: np.ndarray, v: Any, op: "TensorSearchsorted"):
        out = op.outputs[0]
        i = out.index[0]
        side = op.side

        raw_v = v
        v = xp.atleast_1d(v)
        searched = xp.searchsorted(a, v, side=op.side)
        xp.add(searched, op.offset, out=searched)
        a_min, a_max = a[0], a[-1]
        if i == 0:
            # the first chunk
            if a_min == a_max:
                miss = v > a_max
            else:
                miss = v > a_max if side == "left" else v >= a_max
        elif i == op.n_chunk - 1:
            # the last chunk
            if a_min == a_max:
                miss = v < a_min
            else:
                miss = v <= a_min if side == "left" else v < a_min
        else:
            if side == "left" and a_min < a_max:
                miss = (v <= a_min) | (v > a_max)
            elif a_min < a_max:
                miss = (v < a_min) | (v >= a_max)
            else:
                assert a_min == a_max
                miss = v != a_min
        if side == "right":
            searched[miss] = -1
        else:
            searched[miss] = op.size + 1

        return searched[0] if np.isscalar(raw_v) else searched

    @classmethod
    def execute(cls, ctx, op):
        a = ctx[op.inputs[0].key]
        v = ctx[op.inputs[1].key] if len(op.inputs) == 2 else op.v

        data = []
        if isinstance(a, tuple):
            data.extend(a)
        else:
            data.append(a)
        if len(op.inputs) == 2:
            data.append(v)

        data, device_id, xp = as_same_device(data, device=op.device, ret_extra=True)

        if isinstance(a, tuple):
            a = data[:2]
        else:
            a = data[0]
        if len(op.inputs) == 2:
            v = data[-1]

        with device(device_id):
            if op.stage is None:
                ret = cls._execute_without_stage(xp, a, v, op)
            else:
                assert op.stage == OperandStage.map
                ret = cls._execute_map(xp, a, v, op)
            ctx[op.outputs[0].key] = ret


def searchsorted(a, v, side="left", sorter=None, combine_size=None):
    """
    Find indices where elements should be inserted to maintain order.

    Find the indices into a sorted tensor `a` such that, if the
    corresponding elements in `v` were inserted before the indices, the
    order of `a` would be preserved.

    Assuming that `a` is sorted:

    ======  ============================
    `side`  returned index `i` satisfies
    ======  ============================
    left    ``a[i-1] < v <= a[i]``
    right   ``a[i-1] <= v < a[i]``
    ======  ============================

    Parameters
    ----------
    a : 1-D array_like
        Input tensor. If `sorter` is None, then it must be sorted in
        ascending order, otherwise `sorter` must be an array of indices
        that sort it.
    v : array_like
        Values to insert into `a`.
    side : {'left', 'right'}, optional
        If 'left', the index of the first suitable location found is given.
        If 'right', return the last such index.  If there is no suitable
        index, return either 0 or N (where N is the length of `a`).
    sorter : 1-D array_like, optional
        Optional tensor of integer indices that sort array a into ascending
        order. They are typically the result of argsort.
    combine_size: int, optional
        The number of chunks to combine.

    Returns
    -------
    indices : tensor of ints
        Array of insertion points with the same shape as `v`.

    See Also
    --------
    sort : Return a sorted copy of a tensor.
    histogram : Produce histogram from 1-D data.

    Notes
    -----
    Binary search is used to find the required insertion points.

    This function is a faster version of the builtin python `bisect.bisect_left`
    (``side='left'``) and `bisect.bisect_right` (``side='right'``) functions,
    which is also vectorized in the `v` argument.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.searchsorted([1,2,3,4,5], 3).execute()
    2
    >>> mt.searchsorted([1,2,3,4,5], 3, side='right').execute()
    3
    >>> mt.searchsorted([1,2,3,4,5], [-10, 10, 2, 3]).execute()
    array([0, 5, 1, 2])

    """

    if (
        not isinstance(a, TENSOR_TYPE)
        and sorter is not None
        and not isinstance(sorter, TENSOR_TYPE)
    ):
        a = astensor(np.asarray(a)[sorter])
    else:
        a = astensor(a)
        if sorter is not None:
            a = a[sorter]

    if a.ndim != 1:
        raise ValueError("`a` should be 1-d tensor")
    if a.issparse():
        # does not support sparse tensor
        raise ValueError("`a` should be a dense tensor")
    if side not in {"left", "right"}:
        raise ValueError(f"'{side}' is an invalid value for keyword 'side'")

    if not np.isscalar(v):
        v = astensor(v)

    op = TensorSearchsorted(
        v=v, side=side, dtype=np.dtype(np.intp), combine_size=combine_size
    )
    return op(a, v)



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List, Union

import pandas as pd

from ...core import ENTITY_TYPE, OutputType
from ...utils import implements
from ..utils import is_dataframe
from .aggregation import DataFrameGroupByAgg
from .custom_aggregation import (
    DataFrameCustomGroupByAggMixin,
    register_custom_groupby_agg_func,
)


@register_custom_groupby_agg_func("nunique")
class DataFrameCustomGroupByNuniqueMixin(DataFrameCustomGroupByAggMixin):
    @classmethod
    def _get_level_indexes(
        cls, op: DataFrameGroupByAgg, data: pd.DataFrame
    ) -> List[int]:
        """
        When group by level, get the level index list.
        Level can be int, level name, or sequence of such.
        This function calculates the corresponding indexes.
        Parameters
        ----------
        op
        data

        Returns
        -------

        """
        index = [data.index.name] if data.index.name else data.index.names
        index = pd.Index(index)
        level = op.groupby_params["level"]
        if isinstance(level, int):
            indexes = [level]
        elif isinstance(level, str):
            indexes = [index.get_loc(level)]
        else:
            level = list(level)
            if isinstance(level[0], int):
                indexes = level
            else:
                indexes = index.get_indexer(level).tolist()
        return indexes

    @classmethod
    def _get_selection_columns(cls, op: DataFrameGroupByAgg) -> Union[None, List]:
        """
        Get groupby selection columns from op parameters.
        If this returns None, it means all columns are required.
        Parameters
        ----------
        op

        Returns
        -------

        """
        if "selection" in op.groupby_params:
            selection = op.groupby_params["selection"]
            if isinstance(selection, (tuple, list)):
                selection = [n for n in selection]
            else:
                selection = [selection]
            return selection

    @classmethod
    def _drop_duplicates_by_series(cls, in_data: pd.DataFrame, origin_cols: List):
        if isinstance(in_data.index, pd.MultiIndex):
            origin_index_name = in_data.index.names
        else:
            origin_index_name = in_data.index.name
        res = in_data.reset_index()
        new_cols = list(res.columns)
        index_cols = [v for v in new_cols if v not in origin_cols]
        res = res.drop_duplicates().set_index(index_cols)
        if isinstance(res.index, pd.MultiIndex):
            res.index.names = origin_index_name
        else:
            res.index.name = origin_index_name
        return res

    @classmethod
    def _get_execute_map_result(
        cls, op: DataFrameGroupByAgg, in_data: pd.DataFrame
    ) -> Union[pd.DataFrame, pd.Series]:
        selections = cls._get_selection_columns(op)
        by_cols = op.raw_groupby_params["by"]
        if by_cols is not None:
            # When `by` some series, the series will be used to determine the groups.
            # We first need to set the index of the data to these series,
            # and then `reset_index` to let these series become data columns.
            # Next bring these columns for `drop_duplicates` and reset these columns to index.
            if isinstance(by_cols, list) and any(
                [isinstance(v, pd.Series) for v in by_cols]
            ):
                origin_cols = list(in_data.columns)
                res = in_data.set_index(by_cols)
                res = cls._drop_duplicates_by_series(res, origin_cols)
            else:
                cols = (
                    [*selections, *by_cols]
                    if selections is not None
                    else in_data.columns
                )
                res = in_data[cols].drop_duplicates(subset=cols).set_index(by_cols)
        else:  # group by level
            selections = selections if selections is not None else in_data.columns
            level_indexes = cls._get_level_indexes(op, in_data)
            in_data = in_data.reset_index()
            index_names = in_data.columns[level_indexes].tolist()
            cols = [*index_names, *selections]
            res = in_data[cols].drop_duplicates().set_index(index_names)

        # if sort=True is specified， sort index when finishing drop_duplicates.
        if op.raw_groupby_params["sort"]:
            res = res.sort_index()

        if op.output_types[0] == OutputType.series:
            res = res.squeeze()

        return res

    @classmethod
    def _get_execute_combine_result(
        cls, op: DataFrameGroupByAgg, in_data: pd.DataFrame
    ) -> Union[pd.DataFrame, pd.Series]:
        by = op.raw_groupby_params["by"]
        if isinstance(by, list) and any([isinstance(v, ENTITY_TYPE) for v in by]):
            # `in_data` may be series when there is index op after groupby
            origin_cols = (
                list(in_data.columns) if is_dataframe(in_data) else [in_data.name]
            )
            res = cls._drop_duplicates_by_series(in_data, origin_cols)
        else:
            # in_data.index.names means MultiIndex (groupby on multi cols)
            index_col = in_data.index.name or in_data.index.names
            res = in_data.reset_index().drop_duplicates().set_index(index_col)
        if op.output_types[0] == OutputType.series:
            res = res.squeeze()
        return res

    @classmethod
    def _get_execute_agg_result(
        cls, op: DataFrameGroupByAgg, in_data: pd.DataFrame
    ) -> Union[pd.DataFrame, pd.Series]:
        groupby_params = op.groupby_params.copy()
        cols = in_data.index.name or in_data.index.names
        by = op.raw_groupby_params["by"]

        if by is not None:
            if isinstance(by, list) and any(
                [isinstance(_by, ENTITY_TYPE) for _by in by]
            ):
                # nothing to do here, just group by level is correct
                pass
            elif op.output_types[0] == OutputType.dataframe:
                groupby_params.pop("level", None)
                groupby_params["by"] = cols
                in_data = in_data.reset_index()
        else:
            # When group by multi levels, we must get the actual all levels from raw_groupby_params,
            # since level field in op.groupby_params is not correct.
            groupby_params["level"] = op.raw_groupby_params["level"]

        # For the tuple usage: .agg(x=('a', 'nunique')), firstly set `as_index=True`.
        # Otherwise, subsequent processing will lose the information about the grouped columns.
        # TODO: This is due to `reduction` functions, but for now, let's keep it simple.
        if op.raw_func is None:
            groupby_params["as_index"] = True
        res = in_data.groupby(**groupby_params).nunique()
        return res

    @classmethod
    @implements(DataFrameCustomGroupByAggMixin.execute_map)
    def execute_map(cls, op, in_data: pd.DataFrame) -> Union[pd.DataFrame, pd.Series]:
        return cls._get_execute_map_result(op, in_data)

    @classmethod
    @implements(DataFrameCustomGroupByAggMixin.execute_combine)
    def execute_combine(
        cls, op, in_data: pd.DataFrame
    ) -> Union[pd.DataFrame, pd.Series]:
        return cls._get_execute_combine_result(op, in_data)

    @classmethod
    @implements(DataFrameCustomGroupByAggMixin.execute_agg)
    def execute_agg(cls, op, in_data: pd.DataFrame) -> Union[pd.DataFrame, pd.Series]:
        return cls._get_execute_agg_result(op, in_data)



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import builtins

from .adapter import MarsEntity, mars_execute
from .data import DataRef


def _get_mars_entity(ref: DataRef) -> MarsEntity:
    mars_entity = getattr(ref.data, "_mars_entity", None)
    if mars_entity is not None:
        return mars_entity
    else:  # pragma: no cover
        raise NotImplementedError(
            f"Unable to execute an instance of {type(ref).__name__} "
        )


def run(obj: DataRef | list[DataRef] | tuple[DataRef], **kwargs) -> None:
    """
    Manually trigger execution.

    Parameters
    ----------
    obj : DataRef or collection of DataRefs
        DataRef or collection of DataRefs to execute.
    """
    refs = []
    if isinstance(obj, DataRef):
        refs.append(obj)
    else:
        refs.extend(obj)

    refs_to_execute = _collect_executable_user_ns_refs(refs)
    for ref in refs:
        if id(ref) not in refs_to_execute and need_to_execute(ref):
            refs_to_execute[id(ref)] = ref

    mars_tileables = [_get_mars_entity(ref) for ref in refs_to_execute.values()]
    if mars_tileables:
        mars_execute(mars_tileables, **kwargs)


def need_to_execute(ref: DataRef) -> bool:
    mars_entity = _get_mars_entity(ref)
    return (
        hasattr(mars_entity, "_executed_sessions")
        and len(getattr(mars_entity, "_executed_sessions")) == 0
    )


def _is_in_final_results(ref: DataRef, results: list[DataRef]):
    mars_entity = _get_mars_entity(ref)
    result_mars_entities = [_get_mars_entity(result) for result in results]
    for result_mars_entity in result_mars_entities:
        stack = [result_mars_entity]
        while stack:
            e = stack.pop()
            if e.key == mars_entity.key:
                return True
            stack.extend(e.inputs)
    return False


def _collect_executable_user_ns_refs(result_refs: list[DataRef]) -> dict[int, DataRef]:
    """
    Collect DataRefs defined in user's interactive namespace that are able to execute.
    """

    def _is_ipython_output_cache(name: str):
        # _, __, ___ or _<n> where <n> stands for the output line number.
        return name in ["_", "__", "___"] or name.startswith("_") and name[1:].isdigit()

    if not _is_interactive() or not _is_ipython_available():
        return {}

    ipython = getattr(builtins, "get_ipython")()
    return dict(
        (id(v), v)
        for k, v in ipython.user_ns.items()
        if isinstance(v, DataRef)
        and not _is_ipython_output_cache(k)
        and need_to_execute(v)
        and _is_in_final_results(v, result_refs)
    )


def _is_interactive() -> bool:
    import sys

    # See: https://stackoverflow.com/a/64523765/7098025
    return hasattr(sys, "ps1")


def _is_ipython_available() -> bool:
    return (
        hasattr(builtins, "get_ipython")
        and getattr(builtins, "get_ipython", None) is not None
    )



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
from typing import Callable, Dict, Optional

# noinspection PyUnresolvedReferences
from pandas import Timedelta  # noqa: F401
from pandas import DateOffset, Interval, NaT, Timestamp, offsets

try:
    from pandas import NA, NamedAgg  # noqa: F401
except ImportError:  # pragma: no cover
    pass

try:
    from pandas import ArrowDtype  # noqa: F401 pandas >= 1.5.x has this type mapper
except ImportError:
    ArrowDtype = None

from ..core.utils.fallback import unimplemented_func
from . import accessors, core, groupby, plotting, window
from ._config import (
    describe_option,
    get_option,
    option_context,
    reset_option,
    set_eng_float_format,
    set_option,
)
from .core import DataFrame, Index, Series


def _install():
    from .mars_adapters import _install as _install_mars_methods
    from .pandas_adapters import _install as _install_pandas_methods

    _install_mars_methods()
    _install_pandas_methods()


def __dir__():
    from .mars_adapters import MARS_DATAFRAME_CALLABLES
    from .pandas_adapters import collect_pandas_module_members

    return list(MARS_DATAFRAME_CALLABLES.keys()) + list(
        collect_pandas_module_members().keys()
    )


def __getattr__(name: str):
    from .mars_adapters import MARS_DATAFRAME_CALLABLES
    from .pandas_adapters import collect_pandas_module_members

    if name in MARS_DATAFRAME_CALLABLES:
        return MARS_DATAFRAME_CALLABLES[name]
    else:
        import pandas

        if not hasattr(pandas, name):
            raise AttributeError(name)
        elif name in collect_pandas_module_members():
            return collect_pandas_module_members()[name]
        else:  # pragma: no cover
            if inspect.ismethod(getattr(pandas, name)):
                return unimplemented_func
            else:
                raise AttributeError(name)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .fft import TensorFFT, fft
from .fft2 import TensorFFT2, fft2
from .fftfreq import TensorFFTFreq, fftfreq
from .fftn import TensorFFTN, fftn
from .fftshift import TensorFFTShift, fftshift
from .hfft import TensorHFFT, hfft
from .ifft import TensorIFFT, ifft
from .ifft2 import TensorIFFT2, ifft2
from .ifftn import TensorIFFTN, ifftn
from .ifftshift import TensorIFFTShift, ifftshift
from .ihfft import TensorIHFFT, ihfft
from .irfft import TensorIRFFT, irfft
from .irfft2 import TensorIRFFT2, irfft2
from .irfftn import TensorIRFFTN, irfftn
from .rfft import TensorRFFT, rfft
from .rfft2 import TensorRFFT2, rfft2
from .rfftfreq import TensorRFFTFreq, rfftfreq
from .rfftn import TensorRFFTN, rfftn



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import reduce
from typing import Any

import pandas as pd

from ..._mars.config import option_context as mars_option_context
from ..._mars.config import options
from ...core.utils.docstring import attach_module_callable_docstring


def set_option(pat: Any, value: Any) -> None:
    try:
        attr_list = pat.split(".")
        if len(attr_list) == 1:
            getattr(options, attr_list[0])
            setattr(options, attr_list[0], value)
        else:
            setattr(reduce(getattr, attr_list[:-1], options), attr_list[-1], value)
    except:
        pd.set_option(pat, value)


def get_option(pat: Any) -> Any:
    try:
        attr_list = pat.split(".")
        return reduce(getattr, attr_list, options)
    except:
        return pd.get_option(pat)


def reset_option(pat) -> None:
    try:
        options.reset_option(pat)
    except:
        pd.reset_option(pat)


def describe_option(*args, **kwargs) -> None:
    pd.describe_option(*args, **kwargs)


class option_context:
    def __init__(self, *args):
        # convert tuple to dict
        context_dict = dict(args[i : i + 2] for i in range(0, len(args), 2))
        mars_dict = {}
        pd_dict = {}
        for key, value in context_dict.items():
            try:
                pd.get_option(key)
                pd_dict[key] = value
            except:
                mars_dict[key] = value

        self.option_contexts = mars_option_context(mars_dict)
        self.pandas_option_context = pd.option_context(
            *tuple(item for sublist in pd_dict.items() for item in sublist)
        )

    def __enter__(self):
        self.option_contexts.__enter__()
        self.pandas_option_context.__enter__()

    def __exit__(self, type, value, traceback):
        self.option_contexts.__exit__(type, value, traceback)
        self.pandas_option_context.__exit__(type, value, traceback)


def set_eng_float_format(accuracy: int = 3, use_eng_prefix: bool = False) -> None:
    pd.set_eng_float_format(accuracy, use_eng_prefix)


attach_module_callable_docstring(set_option, pd, pd.set_option)

attach_module_callable_docstring(get_option, pd, pd.get_option)

attach_module_callable_docstring(reset_option, pd, pd.reset_option)

attach_module_callable_docstring(describe_option, pd, pd.describe_option)

attach_module_callable_docstring(option_context, pd, pd.option_context)

attach_module_callable_docstring(set_eng_float_format, pd, pd.set_eng_float_format)



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Union

import pandas

from ..core import DataRef
from ..core.adapter import (
    MarsDatetimeAccessor,
    MarsGetAttrProxy,
    MarsStringAccessor,
    register_converter,
    to_mars,
)
from ..core.utils.docstring import attach_module_callable_docstring
from .mars_adapters.core import install_members


@register_converter(from_cls_list=[MarsStringAccessor])
class StringAccessor(MarsGetAttrProxy):
    def __init__(self, obj: Union[DataRef, MarsStringAccessor]):
        if isinstance(obj, MarsStringAccessor):
            # when the owner of the accessor being xorbits.core.data.Data.
            super().__init__(obj)
        elif isinstance(obj, DataRef):
            # when the owner of the accessor being xorbits.pandas.core.Series.
            super().__init__(MarsStringAccessor(to_mars(obj)))


install_members(
    StringAccessor,
    MarsStringAccessor,
    pandas,
    pandas.core.strings.accessor.StringMethods,
)
attach_module_callable_docstring(
    StringAccessor, pandas, pandas.core.strings.accessor.StringMethods
)


@register_converter(from_cls_list=[MarsDatetimeAccessor])
class DatetimeAccessor(MarsGetAttrProxy):
    def __init__(self, obj: Union[DataRef, MarsDatetimeAccessor]):
        if isinstance(obj, MarsDatetimeAccessor):
            # when the owner of the accessor being xorbits.core.data.Data.
            super().__init__(obj)
        elif isinstance(obj, DataRef):
            # when the owner of the accessor being xorbits.pandas.core.Series.
            super().__init__(MarsDatetimeAccessor(to_mars(obj)))


install_members(
    DatetimeAccessor,
    MarsDatetimeAccessor,
    pandas,
    pandas.core.indexes.accessors.CombinedDatetimelikeProperties,
)
attach_module_callable_docstring(
    DatetimeAccessor,
    pandas,
    pandas.core.indexes.accessors.CombinedDatetimelikeProperties,
)
